"""
Tests for operations using GeoPandas.
"""

import json
import math
from pathlib import Path

import geopandas as gpd
import pandas as pd
import pygeoops
import pytest
import shapely.geometry as sh_geom

import geofileops as gfo
from geofileops import GeometryType
from geofileops.util import _geofileinfo, _geometry_util
from geofileops.util import _geoops_gpd as geoops_gpd
from geofileops.util._geofileinfo import GeofileInfo
from tests import test_helper
from tests.test_helper import (
    EPSGS,
    SUFFIXES_GEOOPS,
    TESTFILES,
    WHERE_AREA_GT_5000,
    WHERE_LENGTH_GT_1000,
    WHERE_LENGTH_GT_200000,
    assert_geodataframe_equal,
)


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize(
    "only_geom_input, gridsize, keep_empty_geoms, where_post",
    [
        (False, 0.0, True, "ST_Area({geometrycolumn}) > 70"),
        (True, 0.01, False, None),
    ],
)
def test_apply(
    tmp_path, suffix, only_geom_input, gridsize, keep_empty_geoms, where_post
):
    # Prepare test data
    test_gdf = gpd.GeoDataFrame(
        data=[
            {
                "uidn": 1,
                "min_area": 2,
                "geometry": test_helper.TestData.polygon_small_island,
            },
            {
                "uidn": 2,
                "min_area": 2,
                "geometry": test_helper.TestData.polygon_with_island,
            },
        ],
        crs=31370,
    )
    input_path = tmp_path / f"polygons_small_holes_{suffix}"
    gfo.to_file(test_gdf, input_path)
    output_path = tmp_path / f"{input_path.stem}-output{suffix}"
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Run test
    if only_geom_input:
        func = lambda geom: pygeoops.remove_inner_rings(
            geometry=geom, min_area_to_keep=2, crs=input_layerinfo.crs
        )
    else:
        func = lambda row: pygeoops.remove_inner_rings(
            row.geometry,
            min_area_to_keep=row.min_area,
            crs=input_layerinfo.crs,
        )

    gfo.apply(
        input_path=str(input_path),
        output_path=str(output_path),
        func=func,
        only_geom_input=only_geom_input,
        gridsize=gridsize,
        keep_empty_geoms=keep_empty_geoms,
        where_post=where_post,
        batchsize=batchsize,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert gfo.has_spatial_index(output_path) is exp_spatial_index

    # Read result for some more detailed checks
    output_gdf = gfo.read_file(output_path).sort_values("uidn").reset_index(drop=True)
    output_layerinfo = gfo.get_layerinfo(output_path)

    assert len(output_layerinfo.columns) == len(input_layerinfo.columns)
    assert output_layerinfo.geometrytype == GeometryType.MULTIPOLYGON

    # Number of rows depends on keep_empty_geoms and where_post
    if where_post == "ST_Area({geometrycolumn}) > 70" and keep_empty_geoms:
        assert output_layerinfo.featurecount == input_layerinfo.featurecount - 1
    elif where_post is None and not keep_empty_geoms:
        assert output_layerinfo.featurecount == input_layerinfo.featurecount
    else:
        raise ValueError(f"unsupported where_post in test: {where_post}")

    for row in output_gdf.itertuples():
        cur_geometry = row.geometry
        assert cur_geometry is not None

        # It should be a normal Polygon, but might be wrapped as MultiPolygon
        if isinstance(cur_geometry, sh_geom.MultiPolygon):
            assert len(cur_geometry.geoms) == 1
            cur_geometry = cur_geometry.geoms[0]
        assert isinstance(cur_geometry, sh_geom.Polygon)

        if row.uidn == 1:
            # In the 1st polygon the island must be removed
            assert len(cur_geometry.interiors) == 0
        elif row.uidn == 2:
            # In the 2nd polygon the island is larger, so should be there
            assert len(cur_geometry.interiors) == 1


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize("only_geom_input", [False, True])
@pytest.mark.parametrize("force_output_geometrytype", [None, GeometryType.POLYGON])
def test_apply_None(tmp_path, suffix, only_geom_input, force_output_geometrytype):
    """
    Some tests regarding None geometries.

    The test uses None geometries as input, but is similar to apply resulting in None.
    """
    # Prepare test data
    test_gdf = gpd.GeoDataFrame(
        data=[
            {
                "id": 1,
                "min_area": 2,
                "geometry": test_helper.TestData.polygon_small_island,
            },
            {
                "id": 2,
                "min_area": 2,
                "geometry": test_helper.TestData.polygon_with_island,
            },
            {"id": 3, "min_area": 2, "geometry": None},
        ],
        crs=31370,
    )
    input_path = tmp_path / f"polygons_small_holes_{suffix}"
    gfo.to_file(test_gdf, input_path)
    output_path = tmp_path / f"{input_path.stem}-output{suffix}"
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    if only_geom_input:
        func = lambda geom: pygeoops.remove_inner_rings(
            geometry=geom, min_area_to_keep=2, crs=input_layerinfo.crs
        )
    else:
        func = lambda row: pygeoops.remove_inner_rings(
            row.geometry, min_area_to_keep=row.min_area, crs=input_layerinfo.crs
        )

    gfo.apply(
        input_path=input_path,
        output_path=output_path,
        func=func,
        only_geom_input=only_geom_input,
        force_output_geometrytype=force_output_geometrytype,
        batchsize=batchsize,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert gfo.has_spatial_index(output_path) is exp_spatial_index

    # Read result for some more detailed checks
    output_gdf = gfo.read_file(output_path).sort_values("id").reset_index(drop=True)
    output_layerinfo = gfo.get_layerinfo(output_path)

    assert len(output_layerinfo.columns) == len(input_layerinfo.columns)
    if force_output_geometrytype is None:
        # The first partial file during calculation to be completed has None geometry,
        # so file is created with GEOMETRY type.
        pass
    elif force_output_geometrytype is None or suffix == ".shp":
        assert output_layerinfo.geometrytype == GeometryType.MULTIPOLYGON
    else:
        assert output_layerinfo.geometrytype == GeometryType.POLYGON

    for index in range(2):
        output_geometry = output_gdf["geometry"][index]
        if index == 2:
            assert output_geometry is None
            continue
        else:
            assert output_geometry is not None
        if isinstance(output_geometry, sh_geom.MultiPolygon):
            assert len(output_geometry.geoms) == 1
            output_geometry = output_geometry.geoms[0]
        assert isinstance(output_geometry, sh_geom.Polygon)

        if index == 0:
            # In the 1st polygon the island must be removed
            assert len(output_geometry.interiors) == 0
        elif index == 1:
            # In the 2nd polygon the island is larger, so should be there
            assert len(output_geometry.interiors) == 1


def test_apply_geooperation_invalid_operation(tmp_path):
    with pytest.raises(ValueError, match="operation not supported: INVALID"):
        geoops_gpd._apply_geooperation(
            input_path=test_helper.get_testfile("polygon-parcel"),
            output_path=tmp_path / "output.gpkg",
            operation="INVALID",
            operation_params={},
        )


@pytest.mark.parametrize(
    "suffix, epsg", [(".gpkg", 31370), (".gpkg", 4326), (".shp", 31370)]
)
def test_buffer_styles(tmp_path, suffix, epsg):
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel", suffix=suffix, epsg=epsg)
    output_path = tmp_path / f"{input_path.stem}-output{suffix}"
    input_layerinfo = gfo.get_layerinfo(input_path)
    assert input_layerinfo.crs is not None
    distance = 1
    if input_layerinfo.crs.is_projected is False:
        # 1 degree = 111 km or 111000 m
        distance /= 111000
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Run standard buffer to compare with
    gfo.buffer(
        input_path=str(input_path),
        output_path=str(output_path),
        distance=distance,
        batchsize=batchsize,
    )

    # Read result
    output_gdf = gfo.read_file(output_path)
    assert output_gdf["geometry"][0] is not None
    area_default_buffer = sum(output_gdf.area)

    # Test polygon buffer with square endcaps
    output_path = (
        output_path.parent / f"{output_path.stem}_endcap_join{output_path.suffix}"
    )
    gfo.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=distance,
        endcap_style=_geometry_util.BufferEndCapStyle.SQUARE,
        join_style=_geometry_util.BufferJoinStyle.MITRE,
        batchsize=batchsize,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert input_layerinfo.featurecount == output_layerinfo.featurecount + 1
    assert len(output_layerinfo.columns) == len(input_layerinfo.columns)
    assert output_layerinfo.geometrytype == GeometryType.MULTIPOLYGON

    # Read result for some more detailed checks
    output_gdf = gfo.read_file(output_path)
    assert output_gdf["geometry"][0] is not None
    area_square_buffer = sum(output_gdf.area)
    assert area_square_buffer > area_default_buffer


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize(
    "epsg, gridsize, explodecollections, where_post",
    [
        (31370, 0.01, True, WHERE_LENGTH_GT_1000),
        (31370, 0.01, False, WHERE_LENGTH_GT_200000),
        (31370, 0.01, True, ""),
        (4326, 0.0, False, None),
    ],
)
def test_dissolve_linestrings(
    tmp_path, suffix, epsg, gridsize, explodecollections, where_post
):
    # Prepare test data
    input_path = test_helper.get_testfile(
        "linestring-watercourse", suffix=suffix, epsg=epsg
    )
    output_basepath = tmp_path / f"{input_path.stem}-output{suffix}"
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Dissolve, no groupby
    output_path = (
        output_basepath.parent / f"{output_basepath.stem}_expl{output_basepath.suffix}"
    )
    gfo.dissolve(
        input_path=str(input_path),
        output_path=str(output_path),
        explodecollections=explodecollections,
        gridsize=gridsize,
        where_post=where_post,
        batchsize=batchsize,
    )

    # Check if the result file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert gfo.has_spatial_index(output_path) is exp_spatial_index
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.geometrytype in [
        GeometryType.LINESTRING,
        GeometryType.MULTILINESTRING,
    ]
    assert len(output_layerinfo.columns) >= 0

    if explodecollections:
        if where_post is None or where_post == "":
            assert output_layerinfo.featurecount == 83
        elif where_post == WHERE_LENGTH_GT_1000:
            assert output_layerinfo.featurecount == 13
        else:
            raise ValueError(f"check for where_post {where_post} not implemented")
    else:
        if where_post is None or where_post == "":
            assert output_layerinfo.featurecount == 1
        elif where_post == WHERE_LENGTH_GT_200000:
            assert output_layerinfo.featurecount == 0
            # Output empty, so nothing more to check
            return
        else:
            raise ValueError(f"check for where_post {where_post} not implemented")

    # Check the contents of the result file
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None
    # TODO: add more in depth check of result


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize("epsg", EPSGS)
def test_dissolve_linestrings_groupby(tmp_path, suffix, epsg):
    # Prepare test data
    input_path = test_helper.get_testfile(
        "linestring-watercourse", suffix=suffix, epsg=epsg
    )
    output_basepath = tmp_path / f"{input_path.stem}-output{suffix}"
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Dissolve, groupby, explodecollections=False
    # -------------------------------------------
    output_path = (
        output_basepath.parent
        / f"{output_basepath.stem}_groupby_noexpl{output_basepath.suffix}"
    )
    groupby_columns = ["NiScoDe"]
    gfo.dissolve(
        input_path=str(input_path),
        output_path=str(output_path),
        groupby_columns=groupby_columns,
        explodecollections=False,
        batchsize=batchsize,
    )

    # Check if the result file is correctly created
    assert output_path.exists()
    input_layerinfo = gfo.get_layerinfo(input_path)

    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.featurecount == 26
    assert output_layerinfo.geometrytype is input_layerinfo.geometrytype
    assert len(output_layerinfo.columns) >= 0

    # Now check the contents of the result file
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None
    # TODO: add more in depth check of result


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize("epsg", EPSGS)
def test_dissolve_linestrings_aggcolumns_columns(tmp_path, suffix, epsg):
    # Prepare test data
    input_path = test_helper.get_testfile(
        "linestring-watercourse", suffix=suffix, epsg=epsg
    )
    output_basepath = tmp_path / f"{input_path.stem}-output{suffix}"
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Dissolve, groupby, explodecollections=False
    # -------------------------------------------
    output_path = (
        output_basepath.parent
        / f"{output_basepath.stem}_groupby_noexpl{output_basepath.suffix}"
    )
    # Also play a bit with casing to check case insnsitivity towards input file, but
    # retaining the casing used in the groupby_columns parameter in output.
    groupby_columns = ["NIScode"]
    agg_columns = {
        "columns": [
            {"column": "fid", "agg": "concat", "as": "fid_concat"},
            {"column": "NaaM", "agg": "max", "as": "naam_MAX"},
        ]
    }
    gfo.dissolve(
        input_path=input_path,
        output_path=output_path,
        groupby_columns=groupby_columns,
        agg_columns=agg_columns,
        explodecollections=False,
        batchsize=batchsize,
    )

    # Check if the result file is correctly created
    assert output_path.exists()
    input_layerinfo = gfo.get_layerinfo(input_path)

    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.featurecount == 26
    assert output_layerinfo.geometrytype is input_layerinfo.geometrytype
    assert len(output_layerinfo.columns) == (
        len(groupby_columns) + len(agg_columns["columns"])
    )

    # Now check the contents of the result file
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None

    # Some more default checks for NISCODE 12009
    niscode_idx = output_gdf[output_gdf["NIScode"] == "12009"].index.item()
    fid_concat_result = sorted(output_gdf["fid_concat"][niscode_idx].split(","))
    if _geofileinfo.get_geofileinfo(input_path).is_fid_zerobased:
        assert fid_concat_result == ["38", "42", "44", "54"]
    else:
        assert fid_concat_result == ["39", "43", "45", "55"]
    assert output_gdf["naam_MAX"][niscode_idx] == "Vosbergbeek"


@pytest.mark.parametrize("agg_columns", [{"json": ["fid", "NaaM"]}, {"json": None}])
def test_dissolve_linestrings_aggcolumns_json(tmp_path, agg_columns):
    # Prepare test data
    input_path = test_helper.get_testfile("linestring-watercourse")
    output_basepath = tmp_path / f"{input_path.stem}-output.gpkg"
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Dissolve, groupby, explodecollections=False
    # -------------------------------------------
    output_path = (
        output_basepath.parent
        / f"{output_basepath.stem}_groupby_noexpl{output_basepath.suffix}"
    )
    # Also play a bit with casing to check case insnsitivity towards input file, but
    # retaining the casing used in the groupby_columns parameter in output.
    groupby_columns = ["NIScode"]
    gfo.dissolve(
        input_path=input_path,
        output_path=output_path,
        groupby_columns=groupby_columns,
        agg_columns=agg_columns,
        explodecollections=False,
        batchsize=batchsize,
    )

    # Check if the result file is correctly created
    assert output_path.exists()
    input_layerinfo = gfo.get_layerinfo(input_path)

    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.featurecount == 26
    assert output_layerinfo.geometrytype is input_layerinfo.geometrytype
    assert len(output_layerinfo.columns) == len(groupby_columns) + 1

    # Now check the contents of the result file
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None

    # Some more default checks for NISCODE 12009
    niscode_idx = output_gdf[output_gdf["NIScode"] == "12009"].index.item()
    json_value = json.loads(str(output_gdf["json"][niscode_idx]))

    # Check NAAM
    naam_result = sorted([value["NAAM"] for value in json_value])
    exp = [
        "Duffelse en Rumstse Scheibeek",
        "Grote Nete",
        "Maltaveldenloop",
        "Vosbergbeek",
    ]
    assert naam_result == exp
    fid_result = sorted([str(value["fid_orig"]) for value in json_value])
    if _geofileinfo.get_geofileinfo(input_path).is_fid_zerobased:
        assert fid_result == ["38", "42", "44", "54"]
    else:
        assert fid_result == ["39", "43", "45", "55"]

    # Some specific tests depending on whether all columns asked or not
    if agg_columns["json"] is None:
        # fid_orig is added to json
        assert len(json_value[0]) == len(input_layerinfo.columns) + 1
    else:
        # fid_orig is added to json
        assert len(json_value[0]) == len(agg_columns["json"]) + 1


@pytest.mark.parametrize(
    "suffix, epsg, explode_input, groupby_columns, explode, gridsize, where_post, "
    "expected_featurecount",
    [
        (".gpkg", 31370, False, ["GEWASgroep"], True, 0.0, "", 26),
        (".gpkg", 31370, False, ["GEWASgroep"], True, 0.01, "", 25),
        (".gpkg", 31370, False, ["GEWASGROEP"], False, 0.0, "", 6),
        (".gpkg", 31370, True, ["GEWASGROEP"], False, 0.0, "", 6),
        (".gpkg", 31370, False, ["gewasGROEP"], False, 0.01, WHERE_AREA_GT_5000, 4),
        (".gpkg", 31370, False, ["gewasGROEP"], True, 0.01, WHERE_AREA_GT_5000, 13),
        (".gpkg", 31370, False, [], True, 0.0, None, 24),
        (".gpkg", 31370, False, None, False, 0.0, None, 1),
        (".gpkg", 4326, False, ["GEWASGROEP"], True, 0.0, None, 26),
        (".shp", 31370, False, ["GEWASGROEP"], True, 0.0, None, 26),
        (".shp", 31370, False, [], True, 0.0, None, 24),
    ],
)
def test_dissolve_polygons(
    tmp_path,
    suffix,
    epsg,
    explode_input,
    groupby_columns,
    explode,
    gridsize,
    where_post,
    expected_featurecount,
):
    # Prepare test data
    test_path = test_helper.get_testfile("polygon-parcel", suffix=suffix, epsg=epsg)
    if explode_input:
        # A bug caused in the past that the output was forced to the same type as the
        # input. If input was simple Polygon, this cause invalid output because
        # MultiPolygons were forced to Simple Polygons.
        input_path = tmp_path / f"input{suffix}"
        gfo.copy_layer(
            src=test_path,
            dst=input_path,
            explodecollections=True,
            force_output_geometrytype=GeometryType.POLYGON,
        )
    else:
        input_path = test_path

    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Test dissolve polygons with different options for groupby and explodecollections
    # --------------------------------------------------------------------------------
    groupby = True if (groupby_columns is None or len(groupby_columns) == 0) else False
    name = f"{input_path.stem}_groupby-{groupby}_explode-{explode}_gridsize-{gridsize}"
    output_path = tmp_path / f"{name}{suffix}"
    gfo.dissolve(
        input_path=input_path,
        output_path=output_path,
        groupby_columns=groupby_columns,
        explodecollections=explode,
        gridsize=gridsize,
        where_post=where_post,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Now check if the tmp file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert gfo.has_spatial_index(output_path) is exp_spatial_index
    assert gfo.isvalid(output_path)
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.featurecount == expected_featurecount
    if groupby is True:
        # No groupby -> normally no columns.
        # Shapefile needs at least one column, if no columns: fid
        if suffix == ".shp":
            assert len(output_layerinfo.columns) == 1
        else:
            assert len(output_layerinfo.columns) == 0
    else:
        assert len(output_layerinfo.columns) == len(groupby_columns)

    if not explode or suffix == ".shp":
        # Shapefile always returns MultiPolygon
        assert output_layerinfo.geometrytype == GeometryType.MULTIPOLYGON
    else:
        assert output_layerinfo.geometrytype == GeometryType.POLYGON

    # Now check the contents of the result file
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None

    # Compare result expected values using geopandas
    columns = ["geometry"]
    if groupby_columns is None or len(groupby_columns) == 0:
        expected_gdf = input_gdf[columns].dissolve()
    else:
        groupby_columns_upper = {column.upper(): column for column in groupby_columns}
        columns += list(groupby_columns_upper)
        expected_gdf = (
            input_gdf[columns].dissolve(by=list(groupby_columns_upper)).reset_index()
        ).rename(columns=groupby_columns_upper)
    expected_gdf = test_helper.prepare_expected_result(
        expected_gdf,
        explodecollections=explode,
        gridsize=gridsize,
        keep_empty_geoms=False,
        where_post=where_post,
    )
    expected_path = tmp_path / f"{input_path.stem}_gpd-output{suffix}"
    gfo.to_file(expected_gdf, expected_path)

    # Small differences with the geopandas result are expected, because gfo
    # adds points in the tiling process. So only basic checks possible.
    # output_gdf.geometry = output_gdf.geometry.simplify(0.01)
    # expected_gdf.geometry = expected_gdf.geometry.simplify(0.01)
    # assert_geodataframe_equal(
    #     output_gdf, expected_gdf, promote_to_multi=True, sort_values=True,
    #     normalize=True, check_less_precise=True
    # )
    if suffix != ".shp":
        # Shapefile needs at least one column, if no columns: fid
        assert list(output_gdf.columns) == list(expected_gdf.columns)
    assert len(output_gdf) == len(expected_gdf)
    output_area = output_gdf.geometry.area.sort_values().reset_index(drop=True)
    expected_area = expected_gdf.geometry.area.sort_values().reset_index(drop=True)
    pd.testing.assert_series_equal(output_area, expected_area)


def test_dissolve_empty_onborder_result(tmp_path):
    """
    Test case where the 1st pass of dissolve does not result in any onborder features.
    """
    test_gdf = gpd.GeoDataFrame(
        geometry=[sh_geom.box(5, 5, 10, 10), sh_geom.box(20, 5, 25, 10)], crs=31370
    )
    test_path = tmp_path / "test.gpkg"
    gfo.to_file(test_gdf, test_path)

    output_path = tmp_path / "output.gpkg"
    gfo.dissolve(
        test_path, output_path=output_path, explodecollections=True, batchsize=1
    )

    assert output_path.exists()
    output_gdf = gfo.read_file(output_path)
    assert_geodataframe_equal(test_gdf, output_gdf, normalize=True, sort_values=True)


@pytest.mark.parametrize("explodecollections", [True, False])
@pytest.mark.parametrize("testfile", TESTFILES)
@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
def test_dissolve_emptyfile(tmp_path, testfile, suffix, explodecollections):
    # Prepare test data
    input_path = test_helper.get_testfile(testfile, suffix=suffix, empty=True)

    # Test dissolve polygons with different options for groupby and explodecollections
    output_path = tmp_path / f"{input_path.stem}_dissolve-emptyfile{suffix}"
    if testfile == "polygon-parcel":
        groupby_columns = ["GEWASGROEP"]
        expected_geometrytype = GeometryType.POLYGON
    elif testfile == "linestring-row-trees":
        groupby_columns = ["rowtype"]
        expected_geometrytype = GeometryType.LINESTRING
    elif testfile == "point":
        groupby_columns = ["type"]
        expected_geometrytype = GeometryType.POINT
    else:
        raise ValueError(f"unimplimented testfile: {testfile}")
    if not explodecollections or suffix == ".shp":
        expected_geometrytype = expected_geometrytype.to_multitype

    gfo.dissolve(
        input_path=input_path,
        output_path=output_path,
        explodecollections=explodecollections,
        groupby_columns=groupby_columns,
    )

    # Now check if the tmp file is correctly created
    assert output_path.exists()
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.featurecount == 0
    assert output_layerinfo.geometrytype == expected_geometrytype
    assert list(output_layerinfo.columns) == groupby_columns


@pytest.mark.parametrize("sql_singlethread", [True, False])
@pytest.mark.parametrize(
    "exp_match, invalid_params",
    [
        ("column in groupby_columns not", {"groupby_columns": "NON_EXISTING_COLUMN"}),
        ("input_path doesn't exist: ", {"input_path": Path("nonexisting.abc")}),
        (
            "output_path must not equal input_path",
            {
                "input_path": test_helper.get_testfile("polygon-parcel"),
                "output_path": test_helper.get_testfile("polygon-parcel"),
            },
        ),
        (
            "Dissolve to tiles is not supported for GeometryType.MULTILINESTRING, ",
            {
                "input_path": test_helper.get_testfile("linestring-watercourse"),
                "nb_squarish_tiles": 2,
            },
        ),
        (
            "abc not available in: ",
            {
                "agg_columns": {
                    "columns": [{"column": "abc", "agg": "count", "as": "cba"}]
                }
            },
        ),
    ],
)
def test_dissolve_invalid_params(tmp_path, sql_singlethread, invalid_params, exp_match):
    """
    Test dissolve with some invalid input params.

    Remark: the structure of agg_columns parameter is tested in
      test_parameter_helper.test_validate_agg_columns_invalid.
    """
    # Prepare test data / params
    input_path = test_helper.get_testfile("polygon-parcel")
    groupby_columns = ["GEWASGROEP"]
    nb_squarish_tiles = 1
    agg_columns = None
    output_path = tmp_path / "output.gpkg"
    for invalid_param in invalid_params:
        if invalid_param == "input_path":
            input_path = invalid_params[invalid_param]
        elif invalid_param == "groupby_columns":
            groupby_columns = invalid_params[invalid_param]
        elif invalid_param == "nb_squarish_tiles":
            nb_squarish_tiles = invalid_params[invalid_param]
        elif invalid_param == "agg_columns":
            agg_columns = invalid_params[invalid_param]
        elif invalid_param == "output_path":
            output_path = invalid_params[invalid_param]
        else:
            ValueError(f"unsupported invalid_param: {invalid_param}")

    # Run test
    with pytest.raises(ValueError, match=exp_match):
        if sql_singlethread:
            if nb_squarish_tiles > 1:
                pytest.skip("nb_squarish_tiles not relevant for dissolve_singlethread")
            from geofileops.util import _geoops_sql

            _geoops_sql.dissolve_singlethread(
                input_path=input_path,
                output_path=output_path,
                groupby_columns=groupby_columns,
                explodecollections=True,
                agg_columns=agg_columns,
                gridsize=0.0,
            )
        else:
            gfo.dissolve(
                input_path=input_path,
                output_path=output_path,
                groupby_columns=groupby_columns,
                explodecollections=True,
                nb_squarish_tiles=nb_squarish_tiles,
                agg_columns=agg_columns,
            )


def test_dissolve_polygons_groupby_None(tmp_path):
    """
    Test dissolve polygons with a column with None values. There was once an issue
    that the type of the column with None Values always ended up as a REAL column after
    the dissolve/group by instead of the original type.
    """

    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel", dst_dir=tmp_path)
    gfo.add_column(input_path, name="none_values", type=gfo.DataType.TEXT)
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Run test
    output_path = tmp_path / "output.gpkg"
    gfo.dissolve(
        input_path=input_path,
        output_path=output_path,
        groupby_columns="none_values",
        explodecollections=True,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Now check if the tmp file is correctly created
    assert output_path.exists()
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.geometrytype == GeometryType.POLYGON
    assert (
        output_layerinfo.columns["none_values"].gdal_type
        == input_layerinfo.columns["none_values"].gdal_type
    )


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
def test_dissolve_polygons_specialcases(tmp_path, suffix):
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel", suffix=suffix)
    output_basepath = tmp_path / f"{input_path.stem}-output{suffix}"
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Test dissolve polygons with specified output layer
    # --------------------------------------------------
    # A different output layer is not supported for shapefile!!!
    output_path = (
        output_basepath.parent
        / f"{output_basepath.stem}_group_outputlayer{output_basepath.suffix}"
    )
    try:
        gfo.dissolve(
            input_path=input_path,
            output_path=output_path,
            groupby_columns=["GEWASGROEP"],
            output_layer="banana",
            explodecollections=True,
            nb_parallel=2,
            batchsize=batchsize,
        )
    except Exception:
        # A different output_layer is not supported for shapefile, so normal
        # that an exception is thrown!
        assert output_path.suffix.lower() == ".shp"

    # Now check if the tmp file is correctly created
    if output_path.suffix.lower() != ".shp":
        assert output_path.exists()
        output_layerinfo = gfo.get_layerinfo(output_path)
        assert output_layerinfo.featurecount == 26
        assert len(output_layerinfo.columns) == 1
        assert output_layerinfo.name == "banana"
        assert output_layerinfo.geometrytype == GeometryType.POLYGON

        # Now check the contents of the result file
        input_gdf = gfo.read_file(input_path)
        output_gdf = gfo.read_file(output_path)
        assert input_gdf.crs == output_gdf.crs
        assert len(output_gdf) == output_layerinfo.featurecount
        assert output_gdf["geometry"][0] is not None

    # Test dissolve to existing output path without and without force
    # ---------------------------------------------------------------
    for force in [True, False]:
        assert output_path.exists() is True
        mtime_orig = output_path.stat().st_mtime
        gfo.dissolve(
            input_path=input_path,
            output_path=output_path,
            explodecollections=True,
            nb_parallel=2,
            batchsize=batchsize,
            force=force,
        )
        if force is False:
            assert output_path.stat().st_mtime == mtime_orig
        else:
            assert output_path.stat().st_mtime != mtime_orig


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize("nb_parallel", [-1, 2])
def test_dissolve_polygons_tiles_empty(tmp_path, suffix, nb_parallel):
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel", suffix=suffix)
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    output_path = tmp_path / f"{input_path.stem}-tilesoutput{suffix}"
    tiles_path = tmp_path / "tiles.gpkg"

    # Make the bounds large enough so there are also empty tiles
    bounds = input_layerinfo.total_bounds
    width = bounds[2] - bounds[0]
    bounds = (
        bounds[0] - 1,
        bounds[1] - 1,
        bounds[2] + 1 + width,
        bounds[3] + 1,
    )
    tiles_gdf = gpd.GeoDataFrame(
        geometry=pygeoops.create_grid2(bounds, nb_squarish_tiles=8), crs=31370
    )
    gfo.to_file(tiles_gdf, tiles_path)

    # Test!
    gfo.dissolve(
        input_path=input_path,
        output_path=output_path,
        tiles_path=tiles_path,
        explodecollections=False,
        nb_parallel=nb_parallel,
        batchsize=batchsize,
        force=True,
    )

    # Now check if the result file is correctly created
    assert output_path.exists()
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.featurecount == 4
    if suffix == ".shp":
        # Shapefile always has an FID field
        # but only if there is no other column???
        # TODO: think about whether this should also be the case for geopackage???
        assert len(output_layerinfo.columns) == 1
    else:
        assert len(output_layerinfo.columns) == 1
    assert output_layerinfo.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
def test_dissolve_polygons_aggcolumns_columns(tmp_path, suffix):
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel", suffix=suffix)
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    output_basepath = tmp_path / f"{input_path.stem}-output{suffix}"

    # Test dissolve polygons with groupby + agg_columns to columns
    output_path = (
        output_basepath.parent
        / f"{output_basepath.stem}_group_aggcolumns{output_basepath.suffix}"
    )
    # Remarks:
    #     - column names are shortened so it also works for shapefile!
    #     - the columns for agg_columns are choosen so they do not contain
    #       unique values, to be a better test case!
    agg_columns = {
        "columns": [
            {"column": "lblHFDtlt", "agg": "max", "as": "lbl_max"},
            {"column": "GEWASGROEP", "agg": "count", "as": "gwsgrp_cnt"},
            {"column": "lblhfdTLT", "agg": "count", "as": "lbl_count"},
            {
                "column": "lblhfdtlt",
                "agg": "count",
                "distinct": True,
                "as": "lbl_cnt_d",
            },
            {"column": "lblhfdtlt", "agg": "concat", "as": "lbl_conc"},
            {
                "column": "lblhfdtlt",
                "agg": "concat",
                "sep": ";",
                "as": "lbl_conc_s",
            },
            {
                "column": "lblhfdtlt",
                "agg": "concat",
                "distinct": True,
                "as": "lbl_conc_d",
            },
            {"column": "hfdtlt", "agg": "mean", "as": "tlt_mea"},
            {"column": "hfdtlt", "agg": "min", "as": "tlt_min"},
            {"column": "hfdtlt", "agg": "sum", "as": "tlt_sum"},
            {"column": "fid", "agg": "concat", "as": "fid_concat"},
        ]
    }
    groupby_columns = ["GEWASgroep"]
    gfo.dissolve(
        input_path=input_path,
        output_path=output_path,
        groupby_columns=groupby_columns,
        agg_columns=agg_columns,
        explodecollections=False,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Now check if the tmp file is correctly created
    assert output_path.exists()
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.featurecount == 6
    assert len(output_layerinfo.columns) == (
        len(groupby_columns) + len(agg_columns["columns"])
    )
    assert output_layerinfo.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None

    # Check agg_columns results
    grasland_idx = output_gdf[output_gdf["GEWASgroep"] == "Grasland"].index.to_list()[0]
    assert output_gdf["lbl_max"][grasland_idx] == "Grasland"
    assert output_gdf["gwsgrp_cnt"][grasland_idx] == 31
    assert output_gdf["lbl_count"][grasland_idx] == 31
    print(f"output_gdf.lbl_concat_distinct: {output_gdf['lbl_conc_d'][grasland_idx]}")
    assert output_gdf["lbl_cnt_d"][grasland_idx] == 1
    assert output_gdf["lbl_conc"][grasland_idx].startswith("Grasland,Grasland,")
    assert output_gdf["lbl_conc_s"][grasland_idx].startswith("Grasland;Grasland;")
    assert output_gdf["lbl_conc_d"][grasland_idx] == "Grasland"
    assert output_gdf["tlt_mea"][grasland_idx] == 60
    assert int(output_gdf["tlt_min"][grasland_idx]) == 60
    assert output_gdf["tlt_sum"][grasland_idx] == 1860

    groenten_idx = output_gdf[
        output_gdf["GEWASgroep"] == "Groenten, kruiden en sierplanten"
    ].index.to_list()[0]
    assert output_gdf["lbl_count"][groenten_idx] == 5
    print(
        "groenten.lblhfdtlt_concat_distinct: "
        "f{output_gdf['lbl_conc_d'][groenten_idx]}"
    )
    assert output_gdf["lbl_cnt_d"][groenten_idx] == 4
    fid_concat_result = sorted(output_gdf["fid_concat"][groenten_idx].split(","))
    if _geofileinfo.get_geofileinfo(input_path).is_fid_zerobased:
        assert fid_concat_result == ["41", "42", "43", "44", "45"]
    else:
        assert fid_concat_result == ["42", "43", "44", "45", "46"]


@pytest.mark.parametrize(
    "agg_columns", [{"json": ["lengte", "oppervl", "lblhfdtlt"]}, {"json": None}]
)
def test_dissolve_polygons_aggcolumns_json(tmp_path, agg_columns):
    # In shapefiles, the length of str columns is very limited, so the json
    # test would fail.
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel")
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    output_basepath = tmp_path / f"{input_path.stem}-output.gpkg"

    # Test dissolve polygons with groupby + agg_columns to json
    output_path = (
        output_basepath.parent
        / f"{output_basepath.stem}_group_aggjson{output_basepath.suffix}"
    )
    gfo.dissolve(
        input_path=input_path,
        output_path=output_path,
        groupby_columns=["GEWASgroep"],
        agg_columns=agg_columns,
        explodecollections=False,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Now check if the tmp file is correctly created
    assert output_path.exists()
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert output_layerinfo.featurecount == 6
    assert len(output_layerinfo.columns) == 2
    assert output_layerinfo.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None
    grasland_json = json.loads(str(output_gdf["json"][0]))
    assert len(grasland_json) == 31
    grasland_json_firstrow = json.loads(str(grasland_json[0]))
    if agg_columns["json"] is None:
        # fid_orig column is added in json, but index column disappeared ???
        assert len(grasland_json_firstrow) == len(input_layerinfo.columns)
    else:
        # fid_orig column is added in json
        assert len(grasland_json_firstrow) == len(agg_columns["json"]) + 1


@pytest.mark.parametrize(
    "suffix, epsg, testfile, gridsize",
    [
        (".gpkg", 31370, "polygon-parcel", 0.01),
        (".gpkg", 31370, "linestring-row-trees", 0.01),
        (".gpkg", 4326, "polygon-parcel", 0.0),
        (".shp", 31370, "polygon-parcel", 0.01),
        (".shp", 4326, "polygon-parcel", 0.0),
    ],
)
def test_simplify_lang(tmp_path, suffix, epsg, testfile, gridsize):
    input_path = test_helper.get_testfile(testfile, suffix=suffix, epsg=epsg)
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    assert input_layerinfo.crs is not None
    if input_layerinfo.crs.is_projected:
        tolerance = 5
    else:
        # 1 degree = 111 km or 111000 m
        tolerance = 5 / 111000
    # Test lang algorithm
    output_path = tmp_path / f"{input_path.stem}-output_lang{suffix}"
    gfo.simplify(
        input_path=str(input_path),
        output_path=str(output_path),
        tolerance=tolerance,
        algorithm=_geometry_util.SimplifyAlgorithm.LANG,
        lookahead=8,
        gridsize=gridsize,
        batchsize=batchsize,
    )

    # Check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert gfo.has_spatial_index(output_path) is exp_spatial_index
    output_layerinfo = gfo.get_layerinfo(output_path)
    expected_featurecount = input_layerinfo.featurecount
    if testfile == "polygon-parcel":
        # The EMPTY geometry will be removed
        expected_featurecount -= 1
        if gridsize > 0.0:
            # The sliver geometry will be removed
            expected_featurecount -= 1
    assert output_layerinfo.featurecount == expected_featurecount
    assert len(output_layerinfo.columns) == len(input_layerinfo.columns)
    assert output_layerinfo.geometrytype == input_layerinfo.geometrytype

    # Check the contents of the result file
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None
    # TODO: some more in-depth validations would be better


@pytest.mark.parametrize(
    "suffix, epsg, testfile, gridsize",
    [
        (".gpkg", 31370, "polygon-parcel", 0.01),
        (".gpkg", 31370, "linestring-row-trees", 0.01),
        (".gpkg", 4326, "polygon-parcel", 0.0),
        (".shp", 31370, "polygon-parcel", 0.01),
        (".shp", 4326, "polygon-parcel", 0.0),
    ],
)
def test_simplify_vw(tmp_path, suffix, epsg, testfile, gridsize):
    # Skip test if simplification is not available
    _ = pytest.importorskip("simplification")

    # Init
    input_path = test_helper.get_testfile(testfile, suffix=suffix, epsg=epsg)
    input_layerinfo = gfo.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    assert input_layerinfo.crs is not None
    if input_layerinfo.crs.is_projected:
        tolerance = 5
    else:
        # 1 degree = 111 km or 111000 m
        tolerance = 5 / 111000

    # Test vw (visvalingam-whyatt) algorithm
    output_path = tmp_path / f"{input_path.stem}-output_vw{suffix}"
    gfo.simplify(
        input_path=input_path,
        output_path=output_path,
        tolerance=tolerance,
        algorithm=_geometry_util.SimplifyAlgorithm.VISVALINGAM_WHYATT,
        gridsize=gridsize,
        batchsize=batchsize,
    )

    # Check if the file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert gfo.has_spatial_index(output_path) is exp_spatial_index
    output_layerinfo = gfo.get_layerinfo(output_path)
    expected_featurecount = input_layerinfo.featurecount
    if testfile == "polygon-parcel":
        # The EMPTY geometry will be removed
        expected_featurecount -= 1
        if gridsize > 0.0:
            # The sliver geometry will be removed
            expected_featurecount -= 1
    assert output_layerinfo.featurecount == expected_featurecount
    assert len(input_layerinfo.columns) == len(output_layerinfo.columns)
    assert output_layerinfo.geometrytype == input_layerinfo.geometrytype

    # Check the contents of the result file
    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert input_gdf.crs == output_gdf.crs
    assert len(output_gdf) == output_layerinfo.featurecount
    assert output_gdf["geometry"][0] is not None
    # TODO: a more in-depth check would be better



"""
Run the benchmarks specified.
"""

from benchmark import benchmarker


def main():
    """
    Run the benchmark function(s).
    """
    functions_to_run = [
        "export_by_location_intersects_complexpoly",
        # "export_by_location_intersects",
        # "join_by_location_intersects",
        # "clip",
        # "intersection",
        # "intersection_complexpoly_agri",
        # "intersection_gridsize",
        # "symmetric_difference_complexpolys_agri",
        # "union",
    ]
    # Run all bechmark functions
    # functions_to_run = None
    benchmarker.run_benchmarks(["benchmarks_geofileops"], functions_to_run)


if __name__ == "__main__":
    main()



"""
Tests for functionalities in ogr_util.
"""

import os

import pytest
from osgeo import gdal

import geofileops as gfo
from geofileops._compat import GDAL_GTE_38
from geofileops.util import _ogr_util
from tests import test_helper


@pytest.mark.parametrize(
    "log_details, error_details",
    [
        ([], []),
        (["Logline1", "Logline2", "ERROR1", "ERROR2"], ["ERROR1", "ERROR2"]),
    ],
)
def test_GDALError(log_details, error_details):
    ex = _ogr_util.GDALError(
        "Error", log_details=log_details, error_details=error_details
    )

    ex_str = str(ex)
    if len(log_details) > 0:
        # The line with only "\n" is dropped
        assert len(ex.log_details) == len(log_details)
        assert len(ex.error_details) == 2
        assert "GDAL CPL_LOG ERRORS" in ex_str
        assert "GDAL CPL_LOG ALL" in ex_str
        for line in log_details:
            assert line in ex_str
        for line in error_details:
            assert line in ex_str
    else:
        assert ex.error_details == []
        assert ex.log_details == []


def test_get_drivers():
    drivers = _ogr_util.get_drivers()
    assert len(drivers) > 0
    assert "GPKG" in drivers
    assert "ESRI Shapefile" in drivers


def test_prepare_gdal_options():
    # Some basic variants that should all be OK
    options_ok = [
        {"LAYER_CREATION.SPATIAL_INDEX": True},
        {"layer_creation.spatial_index": True},
        {" LAYER_CREATION . SPATIAL_INDEX ": True},
    ]
    for option in options_ok:
        prepared = _ogr_util._prepare_gdal_options(option)
        assert prepared["LAYER_CREATION.SPATIAL_INDEX"] == "YES"

        prepared = _ogr_util._prepare_gdal_options(option, split_by_option_type=True)
        assert "LAYER_CREATION" in prepared
        assert prepared["LAYER_CREATION"]["SPATIAL_INDEX"] == "YES"

    # Some more specific cases
    prepared = _ogr_util._prepare_gdal_options({"LAYER_CREATION.SPATIAL_INDEX": False})
    assert prepared["LAYER_CREATION.SPATIAL_INDEX"] == "NO"

    # These options should give an error
    options_nok = [
        {"LAYER_CREATION-SPATIAL_INDEX": True},
        {"NOT_EXISTING_OPTION_TYPE.SPATIAL_INDEX": True},
        {"LAYER_CREATION.SPATIAL_INDEX": True, "layer_creation.spatial_index": False},
    ]
    for option in options_nok:
        try:
            _ = _ogr_util._prepare_gdal_options(option)
            error_raised = False
        except Exception:
            error_raised = True
        assert error_raised is True, f"Error should have been raised for {option}"


def test_read_cpl_log(tmp_path):
    # Prepare test data
    cpl_log_path = tmp_path / "cpl_log.log"
    test_log_lines = [
        "logging line 1\n",
        "ERROR1\n",
        "\nERROR2\n",
        "\0\n",
        "\n",
        "logging line 3\n",
    ]
    with open(cpl_log_path, mode="w") as file:
        file.writelines(test_log_lines)

    # Test
    log_lines, error_lines = _ogr_util.read_cpl_log(cpl_log_path)

    assert len(error_lines) == 2
    # There are two lines with no usefull data, thay will be ignored
    assert len(log_lines) == len(test_log_lines) - 2


def test_set_config_options():
    # Init
    test1_config_notset = "TEST_CONFIG_OPTION_1"
    test2_config_alreadyset = "TEST_CONFIG_OPTION_2"
    test3_config_envset = "TEST_CONFIG_OPTION_3"
    test4_bool_true = "TEST_CONFIG_OPTION_4"
    test5_bool_false = "TEST_CONFIG_OPTION_5"
    test6_int_50 = "TEST_CONFIG_OPTION_6"
    assert gdal.GetConfigOption(test1_config_notset) is None
    assert gdal.GetConfigOption(test2_config_alreadyset) is None
    gdal.SetConfigOption(test2_config_alreadyset, "test2_original_value")
    assert test3_config_envset not in os.environ
    os.environ[test3_config_envset] = "test3_original_env_value"

    # Set config options with context manager
    with _ogr_util.set_config_options(
        {
            test1_config_notset: "test1_context_value",
            test2_config_alreadyset: "test2_context_value",
            test3_config_envset: "test3_context_value",
            test4_bool_true: True,
            test5_bool_false: False,
            test6_int_50: 50,
        }
    ):
        assert gdal.GetConfigOption(test1_config_notset) == "test1_context_value"
        assert gdal.GetConfigOption(test2_config_alreadyset) == "test2_context_value"
        assert gdal.GetConfigOption(test3_config_envset) == "test3_context_value"
        assert gdal.GetConfigOption(test4_bool_true) == "YES"
        assert gdal.GetConfigOption(test5_bool_false) == "NO"
        assert gdal.GetConfigOption(test6_int_50) == "50"

    # The options set with context manager should be gone
    assert gdal.GetConfigOption(test1_config_notset) is None
    # TODO: delete next line + uncomment 2nd if GetConfigOptions is supported
    assert gdal.GetConfigOption(test2_config_alreadyset) is None
    # assert gdal.GetConfigOption(test2_config_alreadyset) == "test2_original_value"
    assert gdal.GetConfigOption(test3_config_envset) == "test3_original_env_value"

    # If option via env is changed, it changes here as well
    os.environ[test3_config_envset] = "test3_new_env_value"
    assert gdal.GetConfigOption(test3_config_envset) == "test3_new_env_value"


def test_vector_translate_input_nolayer(tmp_path):
    input_path = test_helper.get_testfile("polygon-parcel", dst_dir=tmp_path)
    output_path = tmp_path / f"output{input_path.suffix}"
    layer = gfo.get_only_layer(input_path)
    gfo.execute_sql(input_path, sql_stmt=f'DROP TABLE "{layer}"')

    with pytest.raises(
        Exception, match="Error .* not recognized as .*a supported file format"
    ):
        _ogr_util.vector_translate(str(input_path), output_path)


@pytest.mark.parametrize(
    "expected_error, kwargs",
    [
        (
            "it is not supported to specify both sql_stmt and where",
            {"where": "abc", "sql_stmt": "def"},
        ),
        (
            "Error input_layers .* not found in",
            {"input_layers": "unexisting"},
        ),
    ],
)
def test_vector_translate_invalid_params(tmp_path, kwargs, expected_error):
    input_path = test_helper.get_testfile("polygon-parcel")
    output_path = tmp_path / f"output{input_path.suffix}"

    with pytest.raises(Exception, match=expected_error):
        _ogr_util.vector_translate(str(input_path), output_path, **kwargs)


@pytest.mark.parametrize("input_suffix", test_helper.SUFFIXES_GEOOPS)
@pytest.mark.parametrize("output_suffix", test_helper.SUFFIXES_GEOOPS)
def test_vector_translate_sql(tmp_path, input_suffix, output_suffix):
    input_path = test_helper.get_testfile("polygon-parcel", suffix=input_suffix)
    output_path = tmp_path / f"output{output_suffix}"
    layer = gfo.get_only_layer(input_path)
    sql_stmt = f'SELECT * FROM "{layer}"'
    _ogr_util.vector_translate(input_path, output_path, sql_stmt=sql_stmt)

    assert output_path.exists()
    input_layerinfo = gfo.get_layerinfo(input_path)
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert len(input_layerinfo.columns) == len(output_layerinfo.columns)
    assert input_layerinfo.featurecount == input_layerinfo.featurecount


@pytest.mark.parametrize(
    "input_suffix, output_suffix, geom_null_asc,"
    "exp_null_geoms_st38, exp_null_geoms_gte38",
    [
        (".gpkg", ".gpkg", True, 9, 9),
        (".gpkg", ".shp", False, 48, 9),
        (".shp", ".gpkg", False, 48, 9),
        (".shp", ".shp", True, 9, 9),
    ],
)
def test_vector_translate_sql_geom_null(
    tmp_path,
    input_suffix,
    output_suffix,
    geom_null_asc,
    exp_null_geoms_st38,
    exp_null_geoms_gte38,
):
    """
    If there the first row of the result has a NULL geometry in the result, all
    geometries become NULL. Using ORDER BY geom IS NULL controls this.
       -> https://github.com/geofileops/geofileops/issues/308
    Has been partly solved in gdal 3.8: for SQLITE query on geopackage it is OK, on a
    .shp file the problem is still there.
    """
    # Prepare test file
    input_path = test_helper.get_testfile("polygon-parcel", suffix=input_suffix)
    output_path = tmp_path / f"output{output_suffix}"
    input_layerinfo = gfo.get_layerinfo(input_path)

    exp_null_geoms = (
        exp_null_geoms_gte38
        if GDAL_GTE_38 and input_suffix == ".gpkg"
        else exp_null_geoms_st38
    )

    orderby_direction = "ASC" if geom_null_asc else "DESC"
    sql_stmt = f"""
        SELECT * FROM (
          SELECT ST_Buffer({input_layerinfo.geometrycolumn}, -10) AS geom, oidn, uidn
            FROM "{input_layerinfo.name}"
          )
        ORDER BY geom IS NULL {orderby_direction}
    """
    _ogr_util.vector_translate(
        input_path,
        output_path,
        sql_stmt=sql_stmt,
        sql_dialect="SQLITE",
        force_output_geometrytype=input_layerinfo.geometrytype,
    )

    assert output_path.exists()
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert "geom" not in output_layerinfo.columns
    assert "geometry" not in output_layerinfo.columns
    assert len(output_layerinfo.columns) == 2

    if output_suffix == ".gpkg":
        assert output_layerinfo.geometrycolumn == "geom"
    else:
        assert output_layerinfo.geometrycolumn == "geometry"

    output_gdf = gfo.read_file(output_path)
    assert "geometry" in output_gdf.columns
    assert len(output_gdf.loc[output_gdf.geometry.isna()]) == exp_null_geoms


@pytest.mark.parametrize("input_suffix", test_helper.SUFFIXES_GEOOPS)
@pytest.mark.parametrize("output_suffix", test_helper.SUFFIXES_GEOOPS)
def test_vector_translate_sql_input_empty(tmp_path, input_suffix, output_suffix):
    input_path = test_helper.get_testfile(
        "polygon-parcel", suffix=input_suffix, empty=True
    )
    output_path = tmp_path / f"output{output_suffix}"
    layer = gfo.get_only_layer(input_path)
    sql_stmt = f'SELECT * FROM "{layer}"'
    _ogr_util.vector_translate(input_path, output_path, sql_stmt=sql_stmt)

    assert output_path.exists()
    input_layerinfo = gfo.get_layerinfo(input_path)
    output_layerinfo = gfo.get_layerinfo(output_path)
    assert len(input_layerinfo.columns) == len(output_layerinfo.columns)



"""
Tests for operations using only gdal.
"""

import sys

import pytest

import geofileops as gfo
from geofileops import GeometryType
from geofileops.util._geofileinfo import GeofileInfo
from tests import test_helper
from tests.test_helper import SUFFIXES_GEOOPS


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
def test_clip_by_geometry(tmp_path, suffix):
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel", suffix=suffix)

    # For Geopackage, also test if fid is properly preserved
    preserve_fid = True if suffix == ".gpkg" else False

    # Do operation
    output_path = tmp_path / f"{input_path.stem}-output{suffix}"
    clip_wkt = (
        "Polygon ((156072 196691, 156036 196836, 156326 196927, 156368 196750, "
        "156072 196691))"
    )
    gfo.clip_by_geometry(
        input_path=input_path, output_path=output_path, clip_geometry=clip_wkt
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert gfo.has_spatial_index(output_path) is exp_spatial_index
    layerinfo_orig = gfo.get_layerinfo(input_path)
    layerinfo_output = gfo.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 22
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)
    assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    output_gdf = gfo.read_file(output_path, fid_as_index=preserve_fid)
    assert output_gdf["geometry"].iloc[0] is not None

    # Check if the fid was properly retained if relevant
    if preserve_fid:
        assert output_gdf.iloc[0:2].index.sort_values().tolist() == [1, 10]


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
def test_export_by_bounds(tmp_path, suffix):
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel", suffix=suffix)

    # For Geopackage, also test if fid is properly preserved
    preserve_fid = True if suffix == ".gpkg" else False

    # Do operation
    output_path = tmp_path / f"{input_path.stem}-output{suffix}"
    filter = (156036, 196691, 156368, 196927)
    gfo.export_by_bounds(
        input_path=str(input_path), output_path=str(output_path), bounds=filter
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert gfo.has_spatial_index(output_path) is exp_spatial_index
    layerinfo_orig = gfo.get_layerinfo(input_path)
    layerinfo_output = gfo.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == 25
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)
    assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    output_gdf = gfo.read_file(output_path, fid_as_index=preserve_fid)
    assert output_gdf["geometry"].iloc[0] is not None

    # Check if the fid was properly retained if relevant
    if preserve_fid:
        assert output_gdf.iloc[0:2].index.sort_values().tolist() == [1, 8]


@pytest.mark.xfail(
    sys.platform == "darwin", reason="warp has precision issues on MacOS14 on arm64"
)
def test_warp(tmp_path):
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel")
    input_gdf = gfo.read_file(input_path)
    input_bounds = input_gdf.total_bounds
    warped_min = 0
    warped_max = 10000
    gcps = [
        (input_bounds[0], input_bounds[1], warped_min, warped_min, None),
        (input_bounds[0], input_bounds[3], warped_min, warped_max, None),
        (input_bounds[2], input_bounds[3], warped_max, warped_max, None),
        (input_bounds[2], input_bounds[1], warped_max, warped_min, None),
    ]

    # Test first with existing output path and force=False
    output_path = tmp_path / f"{input_path.stem}-output.gpkg"
    output_path.touch()
    gfo.warp(
        input_path=str(input_path),
        output_path=str(output_path),
        gcps=gcps,
        algorithm="tps",
    )
    assert output_path.exists()
    assert output_path.stat().st_size == 0

    # Test force=True
    gfo.warp(
        input_path=input_path,
        output_path=output_path,
        gcps=gcps,
        algorithm="tps",
        force=True,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert gfo.has_spatial_index(output_path) is exp_spatial_index
    layerinfo_orig = gfo.get_layerinfo(input_path)
    layerinfo_output = gfo.get_layerinfo(output_path)
    assert layerinfo_output.featurecount == layerinfo_orig.featurecount
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)
    assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # Now check the contents of the result file
    output_gdf = gfo.read_file(output_path)
    assert output_gdf["geometry"][0] is not None
    output_bounds = output_gdf.total_bounds
    assert output_bounds[0] >= warped_min
    assert output_bounds[1] >= warped_min
    assert output_bounds[2] <= warped_max
    assert output_bounds[3] <= warped_max



"""
Module containing some utilities regarding io.
"""

import logging
import os
import tempfile
from pathlib import Path
from typing import Optional

import geofileops as gfo


def create_tempdir(base_dirname: str, parent_dir: Optional[Path] = None) -> Path:
    """
    Creates a new tempdir in the default temp location.

    Remark: the temp dir won't be cleaned up automatically!

    Examples:
        - base_dirname="foo" -> /tmp/foo_000001
        - base_dirname="foo/bar" -> /tmp/foo/bar_000001

    Args:
        base_dirname (str): The name the tempdir will start with. The name will be
            suffixed with a number to make the directory name unique. If a "/" is part
            of the base_dirname a subdirectory will be created: e.g. "foo/bar".
        parent_dir (Path, optional): The dir to create the tempdir in. If None, the
            system temp dir is used. Defaults to None.

    Raises:
        Exception: if it wasn't possible to create the temp dir because there
            wasn't found a unique directory name.

    Returns:
        Path: the path to the temp dir created.
    """
    if parent_dir is None:
        parent_dir = Path(tempfile.gettempdir())

    for i in range(1, 999999):
        try:
            tempdir = parent_dir / f"{base_dirname}_{i:06d}"
            tempdir.mkdir(parents=True)
            return tempdir
        except FileExistsError:
            continue

    raise Exception(
        f"Wasn't able to create a temporary dir with basedir: "
        f"{parent_dir / base_dirname}"
    )


def get_tempfile_locked(
    base_filename: str,
    suffix: str = ".tmp",
    dirname: Optional[str] = None,
    tempdir: Optional[Path] = None,
) -> tuple[Path, Path]:
    """
    Formats a temp file path, and creates a corresponding lock file.

    This way you can treat it immediately as being locked.

    Args:
        base_filename (str): The base filename to use. A numeric suffix will be
            appended to make the filename unique.
        suffix (str, optional): The suffix/extension of the tempfile.
            Defaults to '.tmp'.
        dirname (str, optional): Name of the subdir to put the tempfile in.
            Defaults to None, then the tempfile created is put directly in the
            root of the tempdir.
        tempdir (Path, optional): Root temp dir to create the file in. If no
            tempdir is specified, the default temp dir will be used.
            Defaults to None.

    Raises:
        Exception: if it wasn't possible to create the temp dir because there
            wasn't found a unique file name.

    Returns:
        Tuple[Path, Path]: First path is the temp file, second one is the lock file.
    """
    # If no dir specified, use default temp dir
    if tempdir is None:
        tempdir = Path(tempfile.gettempdir())
    if dirname is not None:
        tempdir = tempdir / dirname
        tempdir.mkdir(parents=True, exist_ok=True)

    # Now look for a unique filename based on the base_filename and put a lock file
    for i in range(1, 999999):
        tempfile_path = tempdir / f"{base_filename}_{i:06d}{suffix}"
        tempfilelock_path = tempdir / f"{base_filename}_{i:06d}{suffix}.lock"
        result = create_file_atomic(tempfilelock_path)
        if result is True:
            if not tempfile_path.exists():
                # OK!
                return (tempfile_path, tempfilelock_path)
            else:
                # Apparently the lock file didn't exist yet, but the file did.
                # So delete lock file and try again.
                tempfilelock_path.unlink()

    raise Exception(
        f"Wasn't able to create a temporary file with base_filename: {base_filename}, "
        f"dir: {dir}"
    )


def create_file_atomic(filename) -> bool:
    """
    Create a lock file in an atomic way, so it is threadsafe.

    Returns True if the file was created by this thread, False if the file existed
    already.
    """
    try:
        fd = os.open(filename, os.O_CREAT | os.O_EXCL)
        os.close(fd)
        return True
    except FileExistsError:
        return False
    except OSError as ex:  # pragma: no cover
        if ex.errno == 13:
            return False
        else:
            raise Exception("Error creating lock file {filename}") from ex


def with_stem(path: Path, new_stem) -> Path:
    # Remark: from python 3.9 this is available on any Path, but to avoid
    # having to require 3.9 for this, this hack...
    return path.parent / f"{new_stem}{path.suffix}"


def output_exists(path: Path, remove_if_exists: bool) -> bool:
    """
    Checks and returns whether the file specified exists.

    If remove_if_exists is True, the file is removed and False is returned.

    Args:
        path (Path): Output file path to check.
        remove_if_exists (bool): If True, remove the output file if it exists.

    Raises:
        ValueError: raised when the output directory does not exist.

    Returns:
        bool: True if the output file exists.
              False if the file didn't exist or if it was removed
              because `remove_if_exists` is True.
    """
    if not path.parent.exists():
        raise ValueError(f"Output directory does not exist: {path.parent}")
    if path.exists():
        if remove_if_exists:
            gfo.remove(path)
            return False
        else:
            logging.info(msg=f"Stop, output exists already {path}", stacklevel=2)
            return True

    return False



"""
Tests for functionalities in _general_util.
"""

import datetime
import os
import time

import geofileops as gfo
from geofileops.util import _general_util


def test_formatbytes():
    bytes_str = _general_util.formatbytes(1)
    assert bytes_str == "1.0 Byte"
    bytes_str = _general_util.formatbytes(2)
    assert bytes_str == "2.0 Bytes"
    bytes_str = _general_util.formatbytes(1024.0)
    assert bytes_str == "1.00 KB"
    bytes_str = _general_util.formatbytes(1024.0**2)
    assert bytes_str == "1.00 MB"
    bytes_str = _general_util.formatbytes(1024.0**3)
    assert bytes_str == "1.00 GB"
    bytes_str = _general_util.formatbytes(1024.0**4)
    assert bytes_str == "1.00 TB"


def test_format_progress():
    start_time = datetime.datetime.now()
    nb_todo = 10000
    for nb_done in range(0, nb_todo + 1, 2000):
        message = _general_util.format_progress(
            start_time=start_time,
            nb_done=nb_done,
            nb_todo=nb_todo,
            operation="test",
            nb_parallel=2,
        )
        time.sleep(0.5)
        if message is not None:
            print(message)


def test_TempEnv():
    test_var1 = "TEST_ENV_VARIABLE1"
    test_var2 = "TEST_ENV_VARIABLE2"
    test_var1_value_context = 1234
    test_var2_value_orig = "test2_value_orig"
    test_var2_value_context = "test2_value_context"

    assert test_var1 not in os.environ
    assert test_var2 not in os.environ
    os.environ[test_var2] = test_var2_value_orig

    with gfo.TempEnv(
        {test_var1: test_var1_value_context, test_var2: test_var2_value_context}
    ):
        assert os.environ[test_var1] == str(test_var1_value_context)
        assert os.environ[test_var2] == test_var2_value_context

    assert test_var1 not in os.environ
    assert os.environ[test_var2] == test_var2_value_orig



"""
Tests for functionalities in ogr_util.
"""

import logging
from pathlib import Path

import pytest

import geofileops as gfo
from geofileops import fileops
from geofileops.util import _sqlite_util as sqlite_util
from geofileops.util._geofileinfo import GeofileInfo
from tests import test_helper
from tests.test_helper import assert_geodataframe_equal


@pytest.mark.parametrize("create_spatial_index", [(True), (False)])
def test_create_table_as_sql(tmp_path, create_spatial_index):
    output_path = tmp_path / "output.gpkg"
    input1_path = test_helper.get_testfile(testfile="polygon-parcel", dst_dir=tmp_path)
    input2_path = test_helper.get_testfile(testfile="polygon-zone", dst_dir=tmp_path)

    sql_stmt = """
            SELECT CastToMulti(ST_CollectionExtract(
                       ST_Intersection(layer1.geom, layer2.geometry), 3)) as geom
                  ,ST_area(layer1.geom) AS area_intersect
                  ,layer1.HFDTLT
                  ,layer2.naam
              FROM {input1_databasename}."parcels" layer1
              JOIN {input1_databasename}."rtree_parcels_geom" layer1tree
                ON layer1.fid = layer1tree.id
              JOIN {input2_databasename}."zones" layer2
              JOIN {input2_databasename}."rtree_zones_geometry" layer2tree
                ON layer2.fid = layer2tree.id
             WHERE 1=1
               AND layer1.rowid > 0 AND layer1.rowid < 10
               AND layer1tree.minx <= layer2tree.maxx
               AND layer1tree.maxx >= layer2tree.minx
               AND layer1tree.miny <= layer2tree.maxy
               AND layer1tree.maxy >= layer2tree.miny
               AND ST_Intersects(layer1.geom, layer2.geometry) = 1
               AND ST_Touches(layer1.geom, layer2.geometry) = 0
            """

    sqlite_util.create_table_as_sql(
        input1_path=input1_path,
        input1_layer="parcels",
        input2_path=input2_path,
        input2_layer="zones",
        output_path=output_path,
        output_layer=output_path.stem,
        output_geometrytype=gfo.GeometryType.MULTIPOLYGON,
        sql_stmt=sql_stmt,
        profile=sqlite_util.SqliteProfile.SPEED,
        create_spatial_index=create_spatial_index,
    )

    assert output_path.exists()
    assert gfo.has_spatial_index(output_path) is create_spatial_index
    output_info = gfo.get_layerinfo(output_path)
    assert output_info.featurecount == 7

    # The gpkg created by spatialite by default include some triggers that have errors
    # and were removed from the gpkg spec but not removed in spatialite.
    # These operations give errors if the triggers are still there.
    gfo.drop_column(output_path, column_name="naam")
    gfo.rename_layer(output_path, layer=output_path.stem, new_layer="test_layername")


@pytest.mark.parametrize(
    "kwargs, expected_error",
    [
        ({"append": True}, "append=True nor update=True are implemented."),
        ({"update": True}, "append=True nor update=True are implemented."),
        (
            {"input1_path": Path("input1.sqlite")},
            "output_path and both input paths must have the same suffix!",
        ),
        (
            {"input2_path": Path("input2.sqlite")},
            "output_path and both input paths must have the same suffix!",
        ),
        (
            {"output_path": Path("output.sqlite")},
            "output_path and both input paths must have the same suffix!",
        ),
    ],
)
def test_create_table_as_sql_invalidparams(kwargs, expected_error):
    # Set default values for kwargs that are not specified:
    if "input1_path" not in kwargs:
        kwargs["input1_path"] = Path("input1.gpkg")
    if "input1_layer" not in kwargs:
        kwargs["input1_layer"] = "input1_layer"
    if "input2_path" not in kwargs:
        kwargs["input2_path"] = Path("input2.gpkg")
    if "input2_layer" not in kwargs:
        kwargs["input2_layer"] = "input2_layer"
    if "output_path" not in kwargs:
        kwargs["output_path"] = Path("output.gpkg")
    if "output_layer" not in kwargs:
        kwargs["output_layer"] = "output_layer"
    if "sql_stmt" not in kwargs:
        kwargs["sql_stmt"] = "SELECTje"
    if "output_geometrytype" not in kwargs:
        kwargs["output_geometrytype"] = None

    with pytest.raises(ValueError, match=expected_error):
        sqlite_util.create_table_as_sql(**kwargs)


def test_execute_sql(tmp_path):
    test_path = test_helper.get_testfile(testfile="polygon-parcel", dst_dir=tmp_path)
    exp_spatial_index = GeofileInfo(test_path).default_spatial_index
    assert gfo.has_spatial_index(test_path) is exp_spatial_index
    info_input = gfo.get_layerinfo(test_path)
    nb_deleted = 0

    # Execute one statement
    sql_stmt = "DELETE FROM parcels WHERE rowid = (SELECT MIN(rowid) FROM parcels)"
    sqlite_util.execute_sql(test_path, sql_stmt=sql_stmt)
    nb_deleted += 1
    info = gfo.get_layerinfo(test_path)
    assert info.featurecount == info_input.featurecount - nb_deleted

    # Execute a list of statements
    sqlite_util.execute_sql(test_path, sql_stmt=[sql_stmt, sql_stmt])
    nb_deleted += 2
    info = gfo.get_layerinfo(test_path)
    assert info.featurecount == info_input.featurecount - nb_deleted


def test_get_columns():
    # Prepare test data
    input1_path = test_helper.get_testfile("polygon-parcel")
    input2_path = test_helper.get_testfile("polygon-zone")

    input1_info = gfo.get_layerinfo(input1_path)
    input2_info = gfo.get_layerinfo(input2_path)
    # Also include an identical column name aliasing a constant, is special case that
    # was a bug (https://github.com/geofileops/geofileops/pull/477).
    sql_stmt = f"""
        SELECT layer1.OIDN, layer1.UIDN, layer1.datum, layer2.naam, 'test' AS naam
          FROM {{input1_databasename}}."{input1_info.name}" layer1
          CROSS JOIN {{input2_databasename}}."{input2_info.name}" layer2
         WHERE 1=1
    """

    # Set logging level to DEBUG so the explain plan logging code path is touched.
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    # Run test
    columns = sqlite_util.get_columns(
        sql_stmt=sql_stmt, input1_path=input1_path, input2_path=input2_path
    )

    assert len(columns) == 5


def test_create_table_as_sql_single_input(tmp_path):
    input_path = test_helper.get_testfile(testfile="polygon-parcel", dst_dir=tmp_path)
    output_path = tmp_path / "output.gpkg"
    distance = 10
    resolution = 5

    # Prepare expected result
    expected_gdf = fileops.read_file(input_path, columns=["HFDTLT"])
    expected_gdf.geometry = expected_gdf.geometry.buffer(
        distance, resolution=resolution
    )

    sql_stmt = f"""
        SELECT ST_buffer(layer.geom, {distance}, {resolution}) as geom
              ,layer.HFDTLT
          FROM "parcels" layer
    """
    sqlite_util.create_table_as_sql(
        input1_path=input_path,
        input1_layer="parcels",
        input2_path=None,
        input2_layer=None,
        output_path=output_path,
        output_layer=None,
        output_geometrytype=None,
        sql_stmt=sql_stmt,
    )

    assert output_path.exists()
    output_gdf = fileops.read_file(output_path)

    # EMPTY geometry becomes NULL/None...
    expected_gdf.loc[expected_gdf.geometry.is_empty, "geometry"] = None
    assert_geodataframe_equal(output_gdf, expected_gdf)



"""
Module to prepare test data for benchmarking geo operations.
"""

import enum
import logging
import pprint
import shutil
import tempfile
import urllib.request
import zipfile
from pathlib import Path
from typing import Optional

import geopandas as gpd
import shapely
import shapely.affinity

import geofileops as gfo

################################################################################
# Some inits
################################################################################

logger = logging.getLogger(__name__)

################################################################################
# The real work
################################################################################


class TestFile(enum.Enum):
    """
    Class to create benchmarking test files.

    Args:
        enum (_type_): _description_

    Raises:
        ValueError: _description_
        RuntimeError: _description_

    Returns:
        _type_: _description_
    """

    AGRIPRC_2018 = (
        0,
        "https://www.landbouwvlaanderen.be/bestanden/gis/Landbouwgebruikspercelen_2018_-_Definitief_(extractie_23-03-2022)_GPKG.zip",
        "agriprc_2018.gpkg",
    )
    AGRIPRC_2019 = (
        1,
        "https://www.landbouwvlaanderen.be/bestanden/gis/Landbouwgebruikspercelen_2019_-_Definitief_(extractie_20-03-2020)_GPKG.zip",
        "agriprc_2019.gpkg",
    )
    COMPLEX_POLYS = (2, None, "complexpolys.gpkg")

    def __init__(self, value, url, filename):
        """
        Create a test file.

        Args:
            value (_type_): _description_
            url (_type_): _description_
            filename (_type_): _description_
        """
        self._value_ = value
        self.url = url
        self.filename = filename

    def get_file(self, output_dir: Path, nb_points: int = 20_000) -> tuple[Path, str]:
        """
        Creates the test file.

        Args:
            output_dir (Path): the directory to write the file to.
            nb_points (int): indication of the number of points the complex polygons
                should consist of. Defaults to 20.000.

        Returns:
            _type_: The path to the file + a description of the test file.
        """
        if self.name != "COMPLEX_POLYS" and nb_points != 20_000:
            raise ValueError("specifying nb_points is only supported for COMPLEX_POLYS")

        if self.url is not None:
            testfile_path = _download_samplefile(
                url=self.url, dst_name=self.filename, dst_dir=output_dir
            )
            testfile_info = gfo.get_layerinfo(testfile_path)
            logger.debug(
                f"TestFile {self.name} contains {testfile_info.featurecount} rows."
            )
            description = f"agri parcels, {testfile_info.featurecount} rows"

        elif self.name == "COMPLEX_POLYS":
            name = Path(self.filename)
            testfile_path = output_dir / f"{name.stem}_{nb_points}{name.suffix}"

            if testfile_path.exists():
                polys_complex_gdf = gpd.read_file(testfile_path, engine="pyogrio")
                nb_coords = shapely.get_num_coordinates(polys_complex_gdf.iloc[0])
                nb_polys = len(polys_complex_gdf)
            else:
                # Prepare some complex polygons to test with
                xmin_start = 30_000
                step = 20_000
                nb_polys = 10
                logger.info(
                    f"create file with {nb_polys} complex polys of ~{nb_points} points"
                )
                poly_complex = _create_complex_poly_points(
                    xmin=xmin_start,
                    ymin=170000.123,
                    width=15000,
                    height=15000,
                    nb_points=nb_points,
                )

                polys_complex = [
                    shapely.affinity.translate(poly_complex, xoff=xoff)
                    for xoff in range(0, (nb_polys * step), step)
                ]
                logger.debug(
                    f"polys_complex: {len(polys_complex)} polys with num_coordinates: "
                    f"{shapely.get_num_coordinates(polys_complex[0])}"
                )
                complex_gdf = gpd.GeoDataFrame(geometry=polys_complex, crs="epsg:31370")
                complex_gdf.to_file(testfile_path, engine="pyogrio")
                nb_coords = shapely.get_num_coordinates(polys_complex[0])
                nb_polys = len(polys_complex)

            description = f"complex polys ({nb_polys} * {nb_coords} coords)"

        else:
            raise RuntimeError(f"get_file not implemented for {self.name}")

        return (testfile_path, description)


def _create_complex_poly_points(
    xmin: float,
    ymin: float,
    width: int,
    height: int,
    nb_points: int,
    nb_points_tol: float = 0.1,
) -> shapely.Polygon:
    if width != 15000 or height != 15000:
        raise ValueError("only width 15000 and height 15000 supported.")
    nb_points_estimate: float = nb_points
    line_distance_estimate = _estimate_line_distance(nb_points)
    nb_points_min = nb_points * (1 - nb_points_tol)
    nb_points_max = nb_points * (1 + nb_points_tol)

    while True:
        poly_complex = _create_complex_poly(
            xmin=xmin,
            ymin=ymin,
            width=width,
            height=height,
            line_distance=line_distance_estimate,
            max_segment_length=100,
        )

        nb_points_created = shapely.get_num_coordinates(poly_complex)
        if nb_points_created < nb_points_min:
            # Not enough points... increase nb_points_estimate
            logger.info(
                f"{nb_points_created=} for {line_distance_estimate=} and "
                f"{nb_points_estimate=} not between {nb_points_min=} and "
                f"{nb_points_max=}"
            )
            nb_points_extra = (nb_points - nb_points_created) / 2
            nb_points_estimate += nb_points_extra
            line_distance_estimate = _estimate_line_distance(nb_points_estimate)
        elif nb_points_created > nb_points_max:
            # Too many points... decrease nb_points_estimate
            logger.info(
                f"{nb_points_created=} for {line_distance_estimate=} and "
                f"{nb_points_estimate=} not between {nb_points_min=} and "
                f"{nb_points_max=}"
            )
            nb_points_less = (nb_points_created - nb_points) / 2
            nb_points_estimate += nb_points_less
            line_distance_estimate = _estimate_line_distance(nb_points_estimate)
        else:
            logger.info(
                f"poly_complex ready with {nb_points_created=} for {nb_points=} "
                f"and {line_distance_estimate=}"
            )

            return poly_complex


def _estimate_line_distance(nb_points: float) -> int:
    # Empirically values found for line_distance and corresponding number of points (for
    # this specific polygon creation function!).
    # Creating a function based on them didn't lead to good results... so use this
    # table and interpolate linearly between the values.
    empirical_values = [
        (85_844, 681),
        (3_007, 3_433),
        (2_017, 7_001),
        (693, 15_242),
        (500, 20_778),
        (250, 50_538),
        (156, 90_325),
        (125, 137_058),
        (105, 192_353),
        (62, 307_842),
        (31, 1_201_306),
    ]

    if nb_points < empirical_values[0][1]:
        raise ValueError(
            f"{nb_points=} not supported, minimal value: {empirical_values[0][1]}"
        )
    if nb_points > empirical_values[-1][1]:
        raise ValueError(
            f"{nb_points=} not supported, maximum value: {empirical_values[-1][1]}"
        )

    distance = -2.0
    for x in range(len(empirical_values) - 1):
        (dist_max, nb_points_min), (dist_min, nb_points_max) = (
            empirical_values[x],
            empirical_values[x + 1],
        )
        if nb_points >= nb_points_min and nb_points <= nb_points_max:
            distance = dist_min + (
                (nb_points - nb_points_min) / (nb_points_max - nb_points_min)
            ) * (dist_max - dist_min)
            break

    if distance < 0:
        raise RuntimeError(f"{nb_points=} gave {distance=}: < 0, so invalid")

    return int(distance)


def _create_complex_poly(
    xmin: float,
    ymin: float,
    width: int,
    height: int,
    line_distance: int,
    max_segment_length: int,
) -> shapely.Polygon:
    """Create complex polygon of a ~grid-shape the size specified."""
    lines = []

    # Vertical lines
    for x_offset in range(0, 0 + width, line_distance):
        lines.append(
            shapely.LineString(
                [(xmin + x_offset, ymin), (xmin + x_offset, ymin + height)]
            )
        )

    # Horizontal lines
    for y_offset in range(0, 0 + height, line_distance):
        lines.append(
            shapely.LineString(
                [(xmin, ymin + y_offset), (xmin + width, ymin + y_offset)]
            )
        )

    poly_complex = shapely.unary_union(shapely.MultiLineString(lines).buffer(2))
    poly_complex = shapely.segmentize(
        poly_complex, max_segment_length=max_segment_length
    )
    assert len(shapely.get_parts(poly_complex)) == 1

    return poly_complex


def _download_samplefile(
    url: str, dst_name: str, dst_dir: Optional[Path] = None
) -> Path:
    """
    Download a sample file to dest_path.

    If it is zipped, it will be unzipped. If needed, it will be converted to
    the file type as determined by the suffix of dst_name.

    Args:
        url (str): the url of the file to download.
        dst_name (str): the name of the destination file.
        dst_dir (Path): the dir to downloaded the sample file to.
            If it is None, a dir in the default tmp location will be
            used. Defaults to None.

    Returns:
        Path: the path to the downloaded sample file.
    """
    # If the destination path is a directory, use the default file name
    dst_path = _prepare_dst_path(dst_name, dst_dir)
    # If the sample file already exists, return
    if dst_path.exists():
        return dst_path
    # Make sure the destination directory exists
    dst_path.parent.mkdir(parents=True, exist_ok=True)

    # If the url points to a file with the same suffix as the dst_path,
    # just download
    url_path = Path(url)
    if url_path.suffix.lower() == dst_path.suffix.lower():
        logger.info(f"Download to {dst_path}")
        urllib.request.urlretrieve(url, dst_path)
    else:
        # The file downloaded is different that the destination wanted, so some
        # converting will need to be done
        tmp_dir = dst_path.parent / "tmp"

        try:
            # Remove tmp dir if it exists already
            if tmp_dir.exists():
                shutil.rmtree(tmp_dir)
            tmp_dir.mkdir(parents=True, exist_ok=True)

            # Download file
            tmp_path = tmp_dir / f"{dst_path.stem}{url_path.suffix.lower()}"
            logger.info(f"Download tmp data to {tmp_path}")
            urllib.request.urlretrieve(url, tmp_path)

            # If the temp file is a .zip file, unzip to dir
            if tmp_path.suffix == ".zip":
                # Unzip
                unzippedzip_dir = dst_path.parent / tmp_path.stem
                logger.info(f"Unzip to {unzippedzip_dir}")
                with zipfile.ZipFile(tmp_path, "r") as zip_ref:
                    zip_ref.extractall(unzippedzip_dir)

                # Look for the file
                tmp_paths = []
                for suffix in [".shp", ".gpkg"]:
                    tmp_paths.extend(list(unzippedzip_dir.rglob(f"*{suffix}")))
                if len(tmp_paths) == 1:
                    tmp_path = tmp_paths[0]
                else:
                    raise Exception(
                        f"Should find 1 geofile, found {len(tmp_paths)}: \n"
                        f"{pprint.pformat(tmp_paths)}"
                    )

            if dst_path.suffix == tmp_path.suffix:
                gfo.move(tmp_path, dst_path)
            else:
                logger.info(f"Convert tmp file to {dst_path}")
                gfo.makevalid(tmp_path, dst_path, keep_empty_geoms=False)
        finally:
            shutil.rmtree(tmp_dir, ignore_errors=True)

    return dst_path


def _prepare_dst_path(dst_name: str, dst_dir: Optional[Path] = None):
    if dst_dir is None:
        return Path(tempfile.gettempdir()) / "geofileops_sampledata" / dst_name
    else:
        return dst_dir / dst_name



"""
Module to generate reports for benchmarks.
"""

import ast
import math
import shutil
import tempfile
from pathlib import Path
from typing import Literal, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pandas.api
from PIL import Image

A4_LONG_SIDE = 11.69
A4_SHORT_SIDE = 8.27


def generate_reports(results_path: Path, output_dir: Path):
    """
    Generate the benchmarking reports.

    Args:
        results_path (Path): the result path to find the results to report on.
        output_dir (Path): directory to write the reports to.
    """
    benchmark_df = pd.read_csv(results_path)

    def format_run_details(input: dict) -> str:
        if input is None or input == np.nan:
            return ""
        if isinstance(input, str):
            input = ast.literal_eval(input)
            result_list = [f"{key}:{input[key]}" for key in input]
            return ";".join(result_list)

        return ""

    # Detailed report per package and per operation
    for package in benchmark_df["package"].unique():
        reports_package_dir = output_dir / package
        reports_package_dir.mkdir(parents=True, exist_ok=True)

        package_df = benchmark_df.loc[benchmark_df["package"] == package]
        for operation in package_df["operation"].unique():
            package_operation_df = package_df.loc[
                benchmark_df["operation"] == operation
            ]
            operation_descr = package_operation_df[
                package_operation_df["run_datetime"]
                == package_operation_df["run_datetime"].max()
            ]["operation_descr"].item()
            package_operation_df = package_operation_df[
                ["package_version", "run_details", "secs_taken"]
            ]
            package_operation_df["run_details"] = package_operation_df[
                "run_details"
            ].apply(lambda x: format_run_details(x))
            package_operation_df = package_operation_df.set_index(
                ["package_version", "run_details"]
            )
            results_report_path = reports_package_dir / f"{package}_{operation}.png"
            save_chart(
                df=package_operation_df,
                title=f"{package}-{operation}\n({operation_descr})",
                size=(8, 6),
                print_labels_on_points=True,
                y_value_formatter="{0:.2f}",
                output_path=results_report_path,
            )

    # Report for last version of each package+operation for comparison
    benchmark_maxversions_df = (
        benchmark_df[["package", "operation", "package_version"]]
        .sort_values(["package", "operation", "package_version"], ascending=False)
        .groupby(["package", "operation"])
        .first()
        .reset_index()
        .set_index(["package", "operation", "package_version"])
    )
    benchmark_maxversion_df = benchmark_df.set_index(
        ["package", "operation", "package_version"]
    )
    benchmark_maxversion_df = (
        benchmark_maxversion_df.loc[
            benchmark_maxversion_df.index.isin(benchmark_maxversions_df.index)
        ].reset_index()
    )[["package", "package_version", "operation", "secs_taken"]]
    benchmark_maxversion_df = benchmark_maxversion_df.pivot_table(
        index="operation", columns=["package", "package_version"]
    )
    # Drop the "secs_taken" level to cleanup legend in chart
    benchmark_maxversion_df = benchmark_maxversion_df.droplevel(level=0, axis=1)
    results_report_path = output_dir / "GeoBenchmark.png"
    save_chart(
        df=benchmark_maxversion_df,
        title="Comparison of libraries, time in sec",
        output_path=results_report_path,
        yscale="log",
        print_labels_on_points=True,
        y_value_formatter="{0:.0f}",
        size=(8, 6),
        linestyle="None",
        gridlines="y",
    )


def save_chart(
    df: pd.DataFrame,
    title: str,
    output_path: Path,
    yscale: Optional[Literal["linear", "log", "symlog", "logit"]] = None,
    y_value_formatter: Optional[str] = None,
    print_labels_on_points: bool = False,
    size: tuple[float, float] = (8, 4),
    plot_kind: Literal[
        "line",
        "bar",
        "barh",
        "hist",
        "box",
        "kde",
        "density",
        "area",
        "pie",
        "scatter",
        "hexbin",
    ] = "line",
    gridlines: Optional[Literal["both", "x", "y"]] = None,
    linestyle: Optional[str] = None,
):
    """
    Render and save a chart.

    Args:
        df (pd.DataFrame): _description_
        title (str): _description_
        output_path (Path): _description_
        yscale (Literal["linear", "log", "symlog", "logit"], optional): y scale to use.
        y_value_formatter (str, optional): a formatter for the y axes and
            labels. Examples:
              - {0:.2%} for a percentage.
              - {0:.2f} for a float with two decimals.
            Defaults to None.
        print_labels_on_points (bool, optional): _description_. Defaults to False.
        size (Tuple[float, float], optional): _description_. Defaults to (8, 4).
        plot_kind (str, optional): _description_. Defaults to "line".
        gridlines (str, optional): where to draw grid lines:

                - 'x': draw grid lines on the x axis
                - 'y': draw grid lines on the x axis
                - 'both': draw grid lines on both axes
            If None, the default for the style used is used. Defaults to None.
        linestyle (Optional[str], optional): _description_. Defaults to None.

    Raises:
        Exception: _description_
    """
    # Init
    # Check input
    non_numeric_columns = [
        column
        for column in df.columns
        if not pandas.api.types.is_numeric_dtype(df[column])
    ]
    if len(non_numeric_columns) > 0:
        raise Exception(
            f"df has non-numeric columns, so cannot be plotted: {non_numeric_columns}"
        )

    # Init some things based on input
    rot = 90

    # Prepare plot figure and axes
    fig, axs = plt.subplots(figsize=(size))
    # Make sure all x axis values are shown
    axs.set_xticks(range(len(df)))
    if yscale is not None:
        plt.yscale(yscale)

    # Plot
    df.plot(ax=axs, kind=plot_kind, rot=rot, title=title, linestyle=linestyle)

    # Show y axes as percentages is asked
    if y_value_formatter is not None:
        axs.yaxis.set_major_formatter(plt.FuncFormatter(y_value_formatter.format))
        axs.yaxis.set_minor_formatter(plt.FuncFormatter(y_value_formatter.format))

    # Show grid lines if specified
    if gridlines is not None:
        axs.grid(axis=gridlines, which="both")

    # Set different markers + print labels
    # Set different markers for each line + get mn/max values + print labels
    markers = ("+", ".", "o", "*")
    max_y_value = None
    min_y_value = None
    for i, line in enumerate(axs.get_lines()):
        line.set_marker(markers[i % len(markers)])

        label_above_line = True
        for index, row in enumerate(df.itertuples()):
            for row_fieldname, row_fieldvalue in row._asdict().items():
                if row_fieldname != "Index":
                    if max_y_value is None or row_fieldvalue > max_y_value:
                        max_y_value = row_fieldvalue
                    if min_y_value is None or row_fieldvalue < min_y_value:
                        min_y_value = row_fieldvalue
                    if print_labels_on_points is True:
                        # Format label
                        if y_value_formatter is not None:
                            text = y_value_formatter.format(row_fieldvalue)
                        else:
                            text = str(row_fieldvalue)

                        # Label below or above line? + switch
                        if label_above_line is True:
                            xytext = (0, 5)
                            label_above_line = False
                        else:
                            xytext = (0, -15)
                            label_above_line = True

                        axs.annotate(
                            text=text,
                            # s=text,
                            # xy=(row.Index, row_fieldvalue),
                            xy=(index, row_fieldvalue),
                            xytext=xytext,
                            textcoords="offset points",
                            ha="center",
                        )

    # Set bottom and top values for y axis
    if max_y_value is not None:
        max_y_value *= 1.1
    if max_y_value is not None and math.isnan(max_y_value) is False:
        plt.ylim(bottom=0, top=max_y_value)
    else:
        plt.ylim(bottom=0)

    # Set legend to the right of the chart
    plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    plt.tight_layout()

    # Save the file if it doesn't exist yet
    if not output_path.exists():
        fig.savefig(str(output_path))
    else:
        # If it exists already, only save it if it has changed
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_output_path = Path(tmp_dir) / output_path.name
            fig.savefig(tmp_output_path)
            img_new = np.asarray(Image.open(tmp_output_path))
            img_old = np.asarray(Image.open(output_path))
            if not np.array_equal(img_new, img_old):
                shutil.move(tmp_output_path, output_path)


if __name__ == "__main__":
    results_path = Path(__file__).resolve().parent / "results/benchmark_results.csv"
    output_dir = Path(__file__).resolve().parent / "results"
    generate_reports(results_path, output_dir)



"""
Setup file to package geofileops.
"""

import setuptools

with open("README.md") as fh:
    long_description = fh.read()

with open("geofileops/version.txt") as file:
    version = file.readline()

setuptools.setup(
    name="geofileops",
    version=version,
    author="Pieter Roggemans",
    author_email="pieter.roggemans@gmail.com",
    description="Python toolbox to process large vector files faster.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/geofileops/geofileops",
    include_package_data=True,
    packages=setuptools.find_packages(),
    install_requires=[
        "cloudpickle",
        "gdal>=3.6,<3.10",
        "geopandas>=0.12,<1",
        "numpy",
        "packaging",
        "pandas",
        "psutil",
        "pygeoops>=0.4,<0.5",
        "pyogrio",
        "pyproj",
        "shapely>=2,<2.1",
    ],
    extras_require={"full": ["simplification"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



"""
Module containing utilities regarding sqlite/spatialite files.
"""

import datetime
import enum
import logging
import pprint
import shutil
import sqlite3
import tempfile
from pathlib import Path
from typing import Optional, Union

from pygeoops import GeometryType

import geofileops as gfo
from geofileops.helpers._configoptions_helper import ConfigOptions
from geofileops.util import _sqlite_userdefined as sqlite_userdefined
from geofileops.util._general_util import MissingRuntimeDependencyError

# Get a logger...
logger = logging.getLogger(__name__)


class EmptyResultError(Exception):
    """
    Exception raised when the SQL statement disn't return any rows.

    Attributes:
        message (str): Exception message
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


def spatialite_version_info() -> dict[str, str]:
    """
    Returns the versions of the spatialite modules.

    Versions returned: spatialite_version, geos_version.

    Raises:
        RuntimeError: if a runtime dependency is not available.

    Returns:
        Dict[str, str]: a dict with the version of the runtime dependencies.
    """
    test_path = Path(__file__).resolve().parent / "test.gpkg"
    conn = sqlite3.connect(test_path)
    try:
        load_spatialite(conn)
        sql = "SELECT spatialite_version(), geos_version()"
        result = conn.execute(sql).fetchall()
        spatialite_version = result[0][0]
        geos_version = result[0][1]
    except MissingRuntimeDependencyError:
        conn.rollback()
        raise
    except Exception as ex:
        conn.rollback()
        raise RuntimeError(f"Error {ex} executing {sql}") from ex
    finally:
        conn.close()

    versions = {
        "spatialite_version": spatialite_version,
        "geos_version": geos_version,
    }
    return versions


class SqliteProfile(enum.Enum):
    DEFAULT = 0
    SPEED = 1


def create_new_spatialdb(path: Path, crs_epsg: Optional[int] = None):
    # Connect to sqlite
    conn = sqlite3.connect(path)
    sql = None
    try:
        with conn:
            load_spatialite(conn)

            # Init file
            output_suffix_lower = path.suffix.lower()
            if output_suffix_lower == ".gpkg":
                sql = "SELECT EnableGpkgMode();"
                # sql = 'SELECT EnableGpkgAmphibiousMode();'
                conn.execute(sql)
                # Remark: this only works on the main database!
                sql = "SELECT gpkgCreateBaseTables();"
                conn.execute(sql)
                if crs_epsg is not None and crs_epsg not in [0, -1, 4326]:
                    sql = f"SELECT gpkgInsertEpsgSRID({crs_epsg})"
                    conn.execute(sql)

                # If they are present, remove triggers that were removed from the gpkg
                # spec because of issues but apparently weren't removed in spatialite.
                # https://github.com/opengeospatial/geopackage/pull/240
                sql = "DROP TRIGGER gpkg_metadata_reference_row_id_value_insert;"
                try:
                    conn.execute(sql)
                except Exception:  # pragma: no cover
                    pass
                sql = "DROP TRIGGER gpkg_metadata_reference_row_id_value_update;"
                try:
                    conn.execute(sql)
                except Exception:  # pragma: no cover
                    pass

            elif output_suffix_lower == ".sqlite":
                sql = "SELECT InitSpatialMetaData(1);"
                conn.execute(sql)
                if crs_epsg is not None and crs_epsg not in [0, -1, 4326]:
                    sql = f"SELECT InsertEpsgSrid({crs_epsg})"
                    conn.execute(sql)
            else:
                raise Exception(f"Unsupported output format: {output_suffix_lower}")

    except Exception as ex:
        raise Exception(f"Error creating spatial db {path}") from ex
    finally:
        conn.close()


def get_columns(
    sql_stmt: str,
    input1_path: Path,
    input2_path: Optional[Path] = None,
    empty_output_ok: bool = True,
    use_spatialite: bool = True,
    output_geometrytype: Optional[GeometryType] = None,
) -> dict[str, str]:
    # Create temp output db to be sure the output DB is writable, even though we only
    # create a temporary table.
    tmp_dir = Path(tempfile.mkdtemp(prefix="geofileops/get_columns_"))
    tmp_path = tmp_dir / f"temp{input1_path.suffix}"
    create_new_spatialdb(path=tmp_path)

    sql = None
    conn = sqlite3.connect(tmp_path, detect_types=sqlite3.PARSE_DECLTYPES)
    try:
        # Load spatialite if asked for
        if use_spatialite:
            load_spatialite(conn)
            if tmp_path.suffix.lower() == ".gpkg":
                sql = "SELECT EnableGpkgMode();"
                conn.execute(sql)

        # Attach to input1
        input1_databasename = "input1"
        sql = f"ATTACH DATABASE ? AS {input1_databasename}"
        dbSpec = (str(input1_path),)
        conn.execute(sql, dbSpec)

        # If input2 isn't the same database input1, attach to it
        input2_databasename = None
        if input2_path is not None:
            if input2_path == input1_path:
                input2_databasename = input1_databasename
            else:
                input2_databasename = "input2"
                sql = f"ATTACH DATABASE ? AS {input2_databasename}"
                dbSpec = (str(input2_path),)
                conn.execute(sql, dbSpec)

        # Prepare sql statement for execute
        sql_stmt_prepared = sql_stmt.format(
            input1_databasename=input1_databasename,
            input2_databasename=input2_databasename,
            batch_filter="",
        )

        # Log explain plan if debug logging enabled.
        if logger.isEnabledFor(logging.DEBUG):
            sql = f"""
                EXPLAIN QUERY PLAN
                SELECT * FROM (
                  {sql_stmt_prepared}
                );
            """
            cur = conn.execute(sql)
            plan = cur.fetchall()
            cur.close()
            logger.debug(pprint.pformat(plan))

        # Create temp table to get the column names + general data types
        # + fetch one row to use it to determine geometrytype.
        # Remark: specify redundant OFFSET 0 to keep sqlite from flattings the subquery.
        sql = f"""
            CREATE TEMPORARY TABLE tmp AS
            SELECT *
              FROM (
                {sql_stmt_prepared}
              )
             LIMIT 1 OFFSET 0;
        """
        conn.execute(sql)
        conn.commit()
        sql = "PRAGMA TABLE_INFO(tmp)"
        cur = conn.execute(sql)
        tmpcolumns = cur.fetchall()
        cur.close()

        # Fetch one row to try to get more detailed data types if needed
        sql = "SELECT * FROM tmp"
        tmpdata = conn.execute(sql).fetchone()
        if tmpdata is not None and len(tmpdata) == 0:
            tmpdata = None
        if not empty_output_ok and tmpdata is None:
            # If no row was returned, stop
            raise EmptyResultError(f"Query didn't return any rows: {sql_stmt}")

        # Loop over all columns to determine the data type
        columns = {}
        for column_index, column in enumerate(tmpcolumns):
            columnname = column[1]
            columntype = column[2]

            if columnname == "geom":
                # PRAGMA TABLE_INFO gives None as column type for a
                # geometry column. So if output_geometrytype not specified,
                # Use ST_GeometryType to get the type
                # based on the data + apply to_multitype to be sure
                if output_geometrytype is None:
                    sql = f'SELECT ST_GeometryType("{columnname}") FROM tmp;'
                    result = conn.execute(sql).fetchall()
                    if len(result) > 0 and result[0][0] is not None:
                        output_geometrytype = GeometryType[result[0][0]].to_multitype
                    else:
                        output_geometrytype = GeometryType["GEOMETRY"]
                columns[columnname] = output_geometrytype.name
            else:
                # If PRAGMA TABLE_INFO doesn't specify the datatype, determine based
                # on data.
                if columntype is None or columntype == "":
                    sql = f'SELECT typeof("{columnname}") FROM tmp;'
                    result = conn.execute(sql).fetchall()
                    if len(result) > 0 and result[0][0] is not None:
                        columns[columnname] = result[0][0]
                    else:
                        # If unknown, take the most general types
                        columns[columnname] = "NUMERIC"
                elif columntype == "NUM":
                    # PRAGMA TABLE_INFO sometimes returns 'NUM', but apparently this
                    # cannot be used in "CREATE TABLE".
                    if tmpdata is None:
                        columns[columnname] = "NUMERIC"
                    elif isinstance(tmpdata[column_index], datetime.date):
                        columns[columnname] = "DATE"
                    elif isinstance(tmpdata[column_index], datetime.datetime):
                        columns[columnname] = "DATETIME"
                    elif isinstance(tmpdata[column_index], str):
                        sql = f'SELECT datetime("{columnname}") FROM tmp;'
                        result = conn.execute(sql).fetchall()
                        if len(result) > 0 and result[0][0] is not None:
                            columns[columnname] = "DATETIME"
                        else:
                            columns[columnname] = "NUMERIC"
                    else:
                        columns[columnname] = "NUMERIC"
                else:
                    columns[columnname] = columntype

    except Exception as ex:
        conn.rollback()
        raise RuntimeError(f"Error {ex} executing {sql}") from ex
    finally:
        conn.close()
        if ConfigOptions.remove_temp_files:
            shutil.rmtree(tmp_dir, ignore_errors=True)

    return columns


def create_table_as_sql(
    input1_path: Path,
    input1_layer: str,
    input2_path: Path,
    input2_layer: str,
    output_path: Path,
    sql_stmt: str,
    output_layer: str,
    output_geometrytype: Optional[GeometryType],
    append: bool = False,
    update: bool = False,
    create_spatial_index: bool = True,
    empty_output_ok: bool = True,
    column_datatypes: Optional[dict] = None,
    profile: SqliteProfile = SqliteProfile.DEFAULT,
):
    """
    Execute sql statement and save the result in the output file.

    Args:
        input1_path (Path): the path to the 1st input file.
        input1_layer (str): the layer/table to select from in het 1st input file
        input2_path (Path): the path to the 2nd input file.
        input2_layer (str): the layer/table to select from in het 2nd input file
        output_path (Path): the path where the output file needs to be created/appended.
        sql_stmt (str): SELECT statement to run on the input files.
        output_layer (str): layer/table name to use.
        output_geometrytype (Optional[GeometryType]): geometry type of the output.
        append (bool, optional): True to append to an existing file. Defaults to False.
        update (bool, optional): True to append to an existing layer. Defaults to False.
        create_spatial_index (bool, optional): True to create a spatial index on the
            output layer. Defaults to True.
        empty_output_ok (bool, optional): If the sql_stmt doesn't return any rows and
            True, create an empty output file. If False, throw EmptyResultError.
            Defaults to True.
        column_datatypes (dict, optional): Can be used to specify the data types of
            columns in the form of {"columnname": "datatype"}. If the data type of
            (some) columns is not specified, it it automatically determined as good as
            possible. Defaults to None.
        profile (SqliteProfile, optional): the set of PRAGMA's to use when creating the
            table. SqliteProfile.DEFAULT will use default setting. SqliteProfile.SPEED
            uses settings optimized for speed, but will be less save regarding
            transaction safety,...
            Defaults to SqliteProfile.DEFAULT.

    Raises:
        ValueError: invalid (combinations of) parameters passed.
        EmptyResultError: the sql_stmt didn't return any rows.
    """
    # Check input parameters
    if append is True or update is True:
        raise ValueError("append=True nor update=True are implemented.")
    output_suffix_lower = output_path.suffix.lower()
    if output_suffix_lower != input1_path.suffix.lower():
        raise ValueError("output_path and both input paths must have the same suffix!")
    if input2_path is not None and output_suffix_lower != input2_path.suffix.lower():
        raise ValueError("output_path and both input paths must have the same suffix!")

    # Check if crs are the same in the input layers + use it (if there is one)
    input1_info = gfo.get_layerinfo(input1_path, input1_layer, raise_on_nogeom=False)
    input2_info = None
    if input2_path is not None:
        input2_info = gfo.get_layerinfo(
            input2_path, input2_layer, raise_on_nogeom=False
        )
    crs_epsg = -1
    if input1_info.crs is not None:
        crs_epsg1 = input1_info.crs.to_epsg()
        if crs_epsg1 is not None:
            crs_epsg = crs_epsg1
        # If input 2 also has a crs, check if it is the same.
        if (
            input2_info is not None
            and input2_info.crs is not None
            and crs_epsg1 != input2_info.crs.to_epsg()
        ):
            logger.warning(
                "input1 layer doesn't have the same crs as input2 layer: "
                f"{input1_info.crs} vs {input2_info.crs}"
            )
    elif input2_info is not None and input2_info.crs is not None:
        crs_epsg2 = input2_info.crs.to_epsg()
        if crs_epsg2 is not None:
            crs_epsg = crs_epsg2

    # If output file doesn't exist yet, create and init it
    if not output_path.exists():
        create_new_spatialdb(path=output_path, crs_epsg=crs_epsg)

    sql = None
    conn = sqlite3.connect(output_path, detect_types=sqlite3.PARSE_DECLTYPES, uri=True)
    try:
        with conn:

            def to_string_for_sql(input) -> str:
                if input is None:
                    return "NULL"
                else:
                    return str(input)

            # Connect to output database file so it is main, otherwise the
            # gpkg... functions don't work
            # Remark: sql statements using knn only work if they are main, so they
            # are executed with ogr, as the output needs to be main as well :-(.
            output_databasename = "main"
            load_spatialite(conn)

            if output_suffix_lower == ".gpkg":
                sql = "SELECT EnableGpkgMode();"
                conn.execute(sql)

            # Set cache size to 128 MB (in kibibytes)
            sql = "PRAGMA cache_size=-128000;"
            conn.execute(sql)
            # Set temp storage to MEMORY
            sql = "PRAGMA temp_store=2;"
            conn.execute(sql)
            # Set soft heap limit to 1 GB (in bytes)
            sql = f"PRAGMA soft_heap_limit={1024*1024*1024};"
            conn.execute(sql)

            # If attach to input1
            input1_databasename = "input1"
            sql = f"ATTACH DATABASE ? AS {input1_databasename}"
            dbSpec = (str(input1_path),)
            # input1_uri = f"file:{input1_path}?immutable=1"
            # dbSpec = (str(input1_uri),)
            conn.execute(sql, dbSpec)

            # If input2 isn't the same database input1, attach to it
            input2_databasename = None
            if input2_path is not None:
                if input2_path == input1_path:
                    input2_databasename = input1_databasename
                else:
                    input2_databasename = "input2"
                    sql = f"ATTACH DATABASE ? AS {input2_databasename}"
                    dbSpec = (str(input2_path),)
                    # input2_uri = f"file:{input1_path}?immutable=1"
                    # dbSpec = (str(input2_uri),)
                    conn.execute(sql, dbSpec)

            # Use the sqlite profile specified
            if profile is SqliteProfile.SPEED:
                # Use memory mapped IO: much faster for calculations
                # (max 30GB)
                conn.execute("PRAGMA mmap_size=30000000000;")

                # These options don't really make a difference on windows, but
                # it doesn't hurt and maybe on other platforms...
                for databasename in [
                    output_databasename,
                    input1_databasename,
                    input2_databasename,
                ]:
                    conn.execute(f"PRAGMA {databasename}.journal_mode=OFF;")

                    # These pragma's increase speed
                    conn.execute(f"PRAGMA {databasename}.locking_mode=EXCLUSIVE;")
                    conn.execute(f"PRAGMA {databasename}.synchronous=OFF;")

            # Determine columns/datatypes to create the table if not specified
            column_types = column_datatypes
            if column_types is None:
                column_types = get_columns(
                    sql_stmt=sql_stmt,
                    input1_path=input1_path,
                    input2_path=input2_path,
                    empty_output_ok=empty_output_ok,
                    use_spatialite=True,
                    output_geometrytype=output_geometrytype,
                )

            # If geometry type was not specified, look for it in column_types
            if output_geometrytype is None and "geom" in column_types:
                output_geometrytype = GeometryType(column_types["geom"])

            # Prepare sql statement
            sql_stmt = sql_stmt.format(
                input1_databasename=input1_databasename,
                input2_databasename=input2_databasename,
            )

            # Now we can create the table
            # Create output table using the gpkgAddGeometryColumn() function
            # Problem: the spatialite function gpkgAddGeometryColumn() doesn't support
            # layer names with special characters (eg. '-')...
            # Solution: mimic the behaviour of gpkgAddGeometryColumn manually.
            # Create table without geom column
            """
            columns_for_create = [
                f'"{columnname}" {column_types[columnname]}\n' for columnname
                in column_types if columnname != 'geom'
            ]
            sql = (
                f'CREATE TABLE {output_databasename}."{output_layer}" '
                f'({", ".join(columns_for_create)})'
            )
            conn.execute(sql)
            # Add geom column with gpkgAddGeometryColumn()
            # Remark: output_geometrytype.name should be detemined from data if needed,
            # see above...
            sql = (
                f"SELECT gpkgAddGeometryColumn("
                f"    '{output_layer}', 'geom', '{output_geometrytype.name}', 0, 0, "
                f"    {to_string_for_sql(crs_epsg)});"
            conn.execute(sql)
            """
            columns_for_create = [
                f'"{columnname}" {column_types[columnname]}\n'
                for columnname in column_types
                if columnname.lower() != "fid"
            ]
            sql = f"""
                CREATE TABLE {output_databasename}."{output_layer}" (
                    fid INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
                    {", ".join(columns_for_create)}
                )
            """
            conn.execute(sql)

            # Add metadata
            if output_suffix_lower == ".gpkg":
                data_type = "features" if "geom" in column_types else "attributes"

                # ~ mimic behaviour of gpkgAddGeometryColumn()
                sql = f"""
                    INSERT INTO {output_databasename}.gpkg_contents (
                        table_name, data_type, identifier, description, last_change,
                        min_x, min_y, max_x, max_y, srs_id)
                    VALUES ('{output_layer}', '{data_type}', NULL, '', DATETIME(),
                        NULL, NULL, NULL, NULL, {to_string_for_sql(crs_epsg)});
                """
                conn.execute(sql)

                # If there is a geometry column, register it
                if "geom" in column_types:
                    assert output_geometrytype is not None
                    sql = f"""
                        INSERT INTO {output_databasename}.gpkg_geometry_columns (
                            table_name, column_name, geometry_type_name, srs_id, z, m)
                        VALUES ('{output_layer}', 'geom', '{output_geometrytype.name}',
                            {to_string_for_sql(crs_epsg)}, 0, 0);
                    """
                    conn.execute(sql)

                    # Now add geom triggers
                    # Remark: this only works on the main database!
                    sql = f"SELECT gpkgAddGeometryTriggers('{output_layer}', 'geom');"
                    conn.execute(sql)
            elif output_suffix_lower == ".sqlite":
                # Create geom metadata if there is one
                if "geom" in column_types:
                    assert output_geometrytype is not None
                    sql = f"""
                        SELECT RecoverGeometryColumn(
                          '{output_layer}', 'geom',
                          {to_string_for_sql(crs_epsg)}, '{output_geometrytype.name}');
                    """
                    conn.execute(sql)

            # Insert data using the sql statement specified
            try:
                columns_for_insert = [f'"{column}"' for column in column_types]
                sql = (
                    f'INSERT INTO {output_databasename}."{output_layer}" '
                    f'({", ".join(columns_for_insert)})\n{sql_stmt}'
                )
                conn.execute(sql)

            except Exception as ex:
                ex_message = str(ex).lower()
                if ex_message.startswith(
                    "unique constraint failed:"
                ) and ex_message.endswith(".fid"):
                    ex.args = (
                        f"{ex}: avoid this by not selecting or aliasing fid "
                        '("select * will select fid!)',
                    )
                raise

            # Create spatial index if needed
            if "geom" in column_types and create_spatial_index is True:
                sql = f"SELECT UpdateLayerStatistics('{output_layer}', 'geom');"
                conn.execute(sql)
                if output_suffix_lower == ".gpkg":
                    # Create the necessary empty index, triggers,...
                    sql = f"SELECT gpkgAddSpatialIndex('{output_layer}', 'geom');"
                    conn.execute(sql)
                    # Now fill the index
                    sql = f"""
                        INSERT INTO "rtree_{output_layer}_geom"
                          SELECT fid
                                ,ST_MinX(geom)
                                ,ST_MaxX(geom)
                                ,ST_MinY(geom)
                                ,ST_MaxY(geom)
                            FROM "{output_layer}"
                           WHERE geom IS NOT NULL
                             AND ST_IsEmpty(geom) = 0
                    """
                    conn.execute(sql)
                elif output_suffix_lower == ".sqlite":
                    sql = f"SELECT CreateSpatialIndex('{output_layer}', 'geom');"
                    conn.execute(sql)
            conn.commit()

    except EmptyResultError:
        logger.info(f"Query didn't return any rows: {sql_stmt}")
        conn.close()
        if output_path.exists():
            output_path.unlink()
    except Exception as ex:
        raise RuntimeError(f"Error {ex} executing {sql}") from ex
    finally:
        if conn is not None:
            conn.close()


def execute_sql(
    path: Path, sql_stmt: Union[str, list[str]], use_spatialite: bool = True
):
    # Connect to database file
    conn = sqlite3.connect(path)
    sql = None

    try:
        if use_spatialite is True:
            load_spatialite(conn)
            if path.suffix.lower() == ".gpkg":
                sql = "SELECT EnableGpkgMode();"
                conn.execute(sql)

        """
        # Set nb KB of cache
        sql = "PRAGMA cache_size=-50000;"
        conn.execute(sql)
        sql = "PRAGMA temp_store=MEMORY;"
        conn.execute(sql)
        conn.execute("PRAGMA journal_mode = WAL")
        """
        if isinstance(sql_stmt, str):
            sql = sql_stmt
            conn.execute(sql)
        else:
            for sql in sql_stmt:
                conn.execute(sql)
        conn.commit()

    except Exception as ex:
        conn.rollback()
        raise Exception(f"Error executing {sql}") from ex
    finally:
        conn.close()


def test_data_integrity(path: Path, use_spatialite: bool = True):
    # Get list of layers in database
    layers = gfo.listlayers(path=path)

    # Connect to database file
    conn = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES)
    sql = None

    try:
        if use_spatialite is True:
            load_spatialite(conn)
        if path.suffix.lower() == ".gpkg":
            sql = "SELECT EnableGpkgMode();"
            conn.execute(sql)

        # Set nb KB of cache
        sql = "PRAGMA cache_size=-50000;"
        conn.execute(sql)
        # Use memory mapped IO = much faster (max 30GB)
        conn.execute("PRAGMA mmap_size=30000000000;")

        # Loop over all layers to check if all data is readable
        for layer in layers:
            sql = f'SELECT * FROM "{layer}"'
            cursor = conn.execute(sql)

            # Fetch the data in chunks to avoid excessive memory usage
            while True:
                result = cursor.fetchmany(10000)
                if not result:
                    # All data was fetched from layer
                    break

    except Exception as ex:
        raise Exception(f"Error executing {sql}") from ex
    finally:
        conn.close()


def load_spatialite(conn):
    """
    Load mod_spatialite for an existing sqlite connection.

    Args:
        conn ([type]): Sqlite connection
    """
    conn.enable_load_extension(True)
    try:
        conn.load_extension("mod_spatialite")
    except Exception as ex:  # pragma: no cover
        raise MissingRuntimeDependencyError(
            "Error trying to load mod_spatialite."
        ) from ex

    # Register custom functions
    conn.create_function(
        "GFO_Difference_Collection",
        -1,
        sqlite_userdefined.gfo_difference_collection,
        deterministic=True,
    )

    conn.create_function(
        "GFO_ReducePrecision",
        -1,
        sqlite_userdefined.gfo_reduceprecision,
        deterministic=True,
    )

    conn.create_function(
        "GFO_Split",
        -1,
        sqlite_userdefined.gfo_split,
        deterministic=True,
    )

    conn.create_function(
        "GFO_Subdivide",
        -1,
        sqlite_userdefined.gfo_subdivide,
        deterministic=True,
    )

    # Register custom aggregate function
    # conn.create_aggregate("GFO_Difference_Agg", 3, userdefined.DifferenceAgg)



"""
Tests for functionalities in sql_util.
"""

from collections.abc import Iterable
from typing import Optional

import pytest

from geofileops.util import _ogr_sql_util


@pytest.mark.parametrize(
    "descr, columns_specified, columns_available, table_alias, columnname_prefix, "
    "fid_column, "
    "exp_quoted, exp_prefixed, exp_prefixed_aliased, "
    "exp_null_aliased, exp_from_subselect",
    [
        (
            "2 columns asked out of 3 in layer",
            ["test1", "TesT2"],
            ["test1", "test3", "test2"],
            "",
            "",
            "fid",
            ',"test1", "test2"',
            ',"test1", "test2"',
            ',"test1" "test1", "test2" "TesT2"',
            ',NULL "test1", NULL "TesT2"',
            ',sub."test1", sub."TesT2"',
        ),
        (
            "fid asked as well as fid_1",
            ["fid", "fid_1", "test2"],
            ["fid_1", "fid_2", "test2"],
            "",
            "",
            "",  # No actual fid column, eg. shapefile
            ',"rowid", "fid_1", "test2"',
            ',"rowid", "fid_1", "test2"',
            ',"rowid" "fid_2", "fid_1" "fid_1", "test2" "test2"',
            ',NULL "fid_2", NULL "fid_1", NULL "test2"',
            ',sub."fid_2", sub."fid_1", sub."test2"',
        ),
        (
            "fid asked and fid_column is fid",
            ["fid", "test1", "test2"],
            ["test2", "test3", "test1"],
            "",
            "",
            "fid",  # Actual" fid" column, eg. geopackage
            ',CAST("fid" AS INT), "test1", "test2"',
            ',CAST("fid" AS INT), "test1", "test2"',
            ',CAST("fid" AS INT) "fid_1", "test1" "test1", "test2" "test2"',
            ',NULL "fid_1", NULL "test1", NULL "test2"',
            ',sub."fid_1", sub."test1", sub."test2"',
        ),
        (
            "fid asked with table alias",
            ["fid", "test1"],
            ["test2", "test3", "test1"],
            "table_alias1",
            "",
            "fid",  # Actual "fid" column, eg. geopackage
            ',CAST("fid" AS INT), "test1"',
            ',CAST(table_alias1."fid" AS INT), table_alias1."test1"',
            ',CAST(table_alias1."fid" AS INT) "fid_1", table_alias1."test1" "test1"',
            ',NULL "fid_1", NULL "test1"',
            ',sub."fid_1", sub."test1"',
        ),
        (
            "fid asked with table alias and prefix",
            ["fid", "test1"],
            ["test2", "test3", "test1"],
            "table_alias1",
            "l1_",
            "ogc_fid",  # Actual fid column named "ogc_fid", eg. sqlite
            ',"ogc_fid", "test1"',
            ',table_alias1."ogc_fid", table_alias1."test1"',
            ',table_alias1."ogc_fid" "l1_fid", table_alias1."test1" "l1_test1"',
            ',NULL "l1_fid", NULL "l1_test1"',
            ',sub."l1_fid", sub."l1_test1"',
        ),
        ("No columns asked", [], ["test1"], "", "", "fid", "", "", "", "", ""),
    ],
)
def test_ColumnFormatter(
    descr: str,
    columns_specified: Optional[list[str]],
    columns_available: Iterable[str],
    table_alias: str,
    columnname_prefix: str,
    fid_column: str,
    exp_quoted: str,
    exp_prefixed: str,
    exp_prefixed_aliased: str,
    exp_null_aliased: str,
    exp_from_subselect: str,
):
    column_frmt = _ogr_sql_util.ColumnFormatter(
        columns_asked=columns_specified,
        columns_in_layer=columns_available,
        fid_column=fid_column,
        table_alias=table_alias,
        column_alias_prefix=columnname_prefix,
    )
    assert column_frmt.quoted() == exp_quoted
    assert column_frmt.prefixed() == exp_prefixed
    assert column_frmt.prefixed_aliased() == exp_prefixed_aliased
    assert column_frmt.null_aliased() == exp_null_aliased
    assert column_frmt.from_subselect() == exp_from_subselect


def test_ColumnFormatter_invalidcolumn():
    with pytest.raises(ValueError, match="columns_asked contains columns not in"):
        _ogr_sql_util.ColumnFormatter(
            columns_asked=["test"],
            columns_in_layer=["test1"],
            fid_column="fid",
        )



import pyogrio
import shapely
from osgeo import gdal
from packaging import version

from geofileops.util import _sqlite_util

PYOGRIO_GTE_07 = version.parse(pyogrio.__version__) >= version.parse("0.7")
SHAPELY_GTE_20 = version.parse(shapely.__version__) >= version.parse("2")

# Check what version of spatialite we are dealing with
spatialite_version_info = _sqlite_util.spatialite_version_info()
spatialite_version = spatialite_version_info["spatialite_version"]
SPATIALITE_GTE_51 = version.parse(spatialite_version) >= version.parse("5.1")

GDAL_GTE_38 = version.parse(gdal.__version__) >= version.parse("3.8")



"""
Module to save layer styles in Geopackage files.
"""

import sqlite3
from pathlib import Path
from typing import Optional

import pandas as pd
from osgeo import gdal, ogr

from geofileops import fileops

gdal.UseExceptions()
ogr.UseExceptions()


def get_layerstyles(
    path: Path, layer: Optional[str] = None, name: Optional[str] = None
) -> pd.DataFrame:
    """
    Get the layer styles saved in the geofile.

    Only styles saved according to the QGIS Geopackage styling extension are read:
    https://github.com/pka/qgpkg/blob/master/qgis_geopackage_extension.md

    Args:
        path (Path): path to the geofile.
        layer (str, optional): the layer to get the styles for. If None, all styles
            regardless of the layer they belong to are returned. Defaults to None.
        name (str, optional): the name of the style to get. If None, all styles
            regardless of their name are returned. Defaults to None.

    Returns:
        pd.DataFrame: the styles found.
    """
    layer_styles_df = fileops.read_file(path, layer="layer_styles", fid_as_index=True)
    layer_styles_df.index.name = "id"
    if layer is not None:
        layer_styles_df = layer_styles_df.loc[layer_styles_df["f_table_name"] == layer]
    if name is not None:
        layer_styles_df = layer_styles_df.loc[layer_styles_df["styleName"] == name]

    return layer_styles_df


def add_layerstyle(
    path: Path,
    layer: str,
    name: str,
    qml: str,
    sld: str = "",
    use_as_default: bool = False,
    description: str = "",
    owner: str = "",
    ui: str = "",
):
    """
    Add the layer style to the geofile.

    Remark: at the time of writing, QGIS only uses the qml field to interprete the
    style, so this field is mandatory and sld is not.

    The style is saved according to the QGIS Geopackage styling extension:
    https://github.com/pka/qgpkg/blob/master/qgis_geopackage_extension.md

    Args:
        path (Path): path to the geofile.
        layer (str): the layer the style is meant for.
        name (str): the name of the style.
        qml (str): the styling in qml format.
        sld (str, optional): the styling in sld format. Defaults to "" as it is not used
            by QGIS at the time of writing.
        use_as_default(bool, optional): True to use the style by default when opening
            the layer in QGIS.
        description (str, optional): description of the style, Defaults to "".
        owner (str, optional): owner of the style, Defaults to "".
        ui (str, optional): ui specification in ui format, Defaults to "".
    """
    # Make sure the layer_styles table exists
    _init_layerstyles(path, exist_ok=True)
    if use_as_default:
        use_as_default_str = 1
    else:
        use_as_default_str = 0

    # Get existing layer styles
    layer_styles_df = get_layerstyles(path, layer=layer, name=name)

    # If the layer style already exists: error
    if len(layer_styles_df) > 0:
        styles_found = (
            layer_styles_df[["f_table_name", "styleName"]]
            .reset_index()
            .to_dict(orient="records")
        )
        raise ValueError(f"layer style exists already: {styles_found}")

    # Insert style
    conn = sqlite3.connect(path)
    sql = """
        INSERT INTO layer_styles (
                id, f_table_catalog, f_table_schema, f_table_name,
                f_geometry_column, styleName, styleQML, styleSLD, useAsDefault,
                description, owner, ui
            )
            VALUES (NULL, '', '', ?, ?, ?, ?, ?, ?, ?, ?, ?)
    """
    try:
        conn.execute(
            sql,
            (layer, "geom", name, qml, sld, use_as_default_str, description, owner, ui),
        )
        conn.commit()
    except Exception as ex:
        conn.rollback()
        raise Exception(f"Error {ex} executing {sql}") from ex
    finally:
        conn.close()


def remove_layerstyle(path: Path, id: int):
    """
    Remove a layer style.

    Only styles saved according to the QGIS Geopackage styling extension are removed:
    https://github.com/pka/qgpkg/blob/master/qgis_geopackage_extension.md

    Args:
        path (Path): path to the geo file.
        id (int): the id of the layer style to remove.
    """
    if not _has_layerstyles_table(path):
        return

    try:
        datasource = gdal.OpenEx(str(path), nOpenFlags=gdal.OF_UPDATE)

        sql = f"DELETE FROM layer_styles WHERE id = {id}"
        result = datasource.ExecuteSQL(sql, dialect="SQLITE")
        datasource.ReleaseResultSet(result)

    finally:
        datasource = None


def _has_layerstyles_table(path: Path) -> bool:
    """
    Check if the layer_styles table exists for the geo file specified.

    Args:
        path (Path): the path to the geofile.

    Returns:
        bool: True of the layer_styles table exists, False if not.
    """
    try:
        datasource = gdal.OpenEx(str(path))
        sql = """
            SELECT count(1) FROM sqlite_master
             WHERE name = 'layer_styles'
               AND type = 'table'
        """
        result = datasource.ExecuteSQL(sql, dialect="SQLITE")
        table_exists = result.GetNextFeature().GetField(0) == 1
        datasource.ReleaseResultSet(result)
    finally:
        datasource = None

    return table_exists


def _init_layerstyles(path: Path, exist_ok: bool = False):
    """
    Create a layer_styles attribute table to store style information in the QGIS way.

    The table is created according to the QGIS Geopackage styling extension:
    https://github.com/pka/qgpkg/blob/master/qgis_geopackage_extension.md

    Args:
        path (Path): the file to create the table in.
        exist_ok (bool, options): If True and the index exists already, don't
            throw an error.
    """
    try:
        # First check if it exists already
        if _has_layerstyles_table(path):
            if exist_ok:
                return
            else:
                raise ValueError(f"layer_styles table exists already in {path}")

        # Doesn't exist yet, so create the table
        datasource = gdal.OpenEx(str(path), nOpenFlags=gdal.OF_UPDATE)
        layer = datasource.CreateLayer(
            "layer_styles", geom_type=ogr.wkbNone, options=["FID=id"]
        )

        # Add the fields we're interested in
        field_name = ogr.FieldDefn("f_table_catalog", ogr.OFTString)
        field_name.SetWidth(256)
        layer.CreateField(field_name)
        field_region = ogr.FieldDefn("f_table_schema", ogr.OFTString)
        field_region.SetWidth(256)
        layer.CreateField(field_region)
        field_region = ogr.FieldDefn("f_table_name", ogr.OFTString)
        field_region.SetWidth(256)
        layer.CreateField(field_region)
        field_region = ogr.FieldDefn("f_geometry_column", ogr.OFTString)
        field_region.SetWidth(256)
        layer.CreateField(field_region)
        field_region = ogr.FieldDefn("styleName", ogr.OFTString)
        field_region.SetWidth(30)
        layer.CreateField(field_region)
        layer.CreateField(ogr.FieldDefn("styleQML", ogr.OFTString))
        layer.CreateField(ogr.FieldDefn("styleSLD", ogr.OFTString))
        layer.CreateField(ogr.FieldDefn("useAsDefault", ogr.OFSTBoolean))
        layer.CreateField(ogr.FieldDefn("description", ogr.OFTString))
        field_region = ogr.FieldDefn("owner", ogr.OFTString)
        field_region.SetWidth(30)
        layer.CreateField(field_region)
        field_region = ogr.FieldDefn("ui", ogr.OFTString)
        field_region.SetWidth(30)
        layer.CreateField(field_region)
        field_region = ogr.FieldDefn("update_time", ogr.OFTDateTime)
        field_region.SetDefault("CURRENT_TIMESTAMP")
        layer.CreateField(field_region)
    finally:
        datasource = None
        layer = None
